import numpy as np

from astropy.convolution import convolve, Gaussian2DKernel
from specutils import Spectrum1D
from specutils.manipulation import gaussian_smooth
from traitlets import List, Unicode, Bool, observe

from jdaviz.core.custom_traitlets import FloatHandleEmpty
from jdaviz.core.events import SnackbarMessage
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import (PluginTemplateMixin, DatasetSelectMixin,
                                        SelectPluginComponent, AddResultsMixin,
                                        with_spinner)
from jdaviz.core.user_api import PluginUserApi

__all__ = ['GaussianSmooth']


@tray_registry('g-gaussian-smooth', label="Gaussian Smooth",
               viewer_requirements=['spectrum', 'flux'])
class GaussianSmooth(PluginTemplateMixin, DatasetSelectMixin, AddResultsMixin):
    """
    See the :ref:`Gaussian Smooth Plugin Documentation <gaussian-smooth>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * ``dataset`` (:class:`~jdaviz.core.template_mixin.DatasetSelect`):
      Dataset to use for computing line statistics.
    * ``mode`` (:class:`~jdaviz.core.template_mixin.SelectPluginComponent`):
      Only available for Cubeviz.  Whether to use spatial or spectral smoothing.
    * :attr:`stddev`:
      Standard deviation of the gaussian to use for smoothing.
    * ``add_results`` (:class:`~jdaviz.core.template_mixin.AddResults`)
    * :meth:`smooth`
    """
    template_file = __file__, "gaussian_smooth.vue"
    stddev = FloatHandleEmpty(1).tag(sync=True)
    show_modes = Bool(False).tag(sync=True)
    mode_items = List().tag(sync=True)
    mode_selected = Unicode().tag(sync=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.config == "cubeviz":
            self.docs_description = 'Smooth data cube spatially or spectrally with a Gaussian kernel.'  # noqa
            self.show_modes = True
            self.dataset._viewers = [
                self._default_flux_viewer_reference_name,
                self._default_spectrum_viewer_reference_name
            ]
            # clear the cache in case the spectrum-viewer selection was already cached
            self.dataset._clear_cache()
        elif self.config in ("mosviz", "specviz2d"):
            # only allow smoothing 1d spectra
            self.dataset._viewers = [self._default_spectrum_viewer_reference_name]
            self.dataset._clear_cache()

        self.dataset.add_filter('not_from_this_plugin')

        self.mode = SelectPluginComponent(self,
                                          items='mode_items',
                                          selected='mode_selected',
                                          manual_options=['Spectral', 'Spatial'])

        # set the filter on the dataset and viewer options
        self._update_dataset_viewer_filters()

        # description displayed under plugin title in tray
        self._plugin_description = 'Smooth data with a Gaussian kernel.'

    @property
    def _default_spectrum_viewer_reference_name(self):
        return getattr(
            self.app._jdaviz_helper, '_default_spectrum_viewer_reference_name', 'spectrum-viewer'
        )

    @property
    def _default_flux_viewer_reference_name(self):
        return getattr(
            self.app._jdaviz_helper, '_default_flux_viewer_reference_name', 'flux-viewer'
        )

    @property
    def user_api(self):
        expose = ['dataset', 'stddev', 'add_results', 'smooth']
        if self.config == "cubeviz":
            expose += ['mode']
        return PluginUserApi(self, expose=expose)

    @observe("dataset_selected", "stddev", "mode_selected")
    def _set_default_results_label(self, event={}):
        '''Generate a label and set the results field to that value'''
        if (hasattr(self, 'dataset') and (len(self.dataset.labels) >= 1) or self.app.config == 'mosviz'):  # noqa
            dataset = f'{self.dataset_selected} '
        else:
            # This should only happen at initialization. Will be overwritten with above
            # once data is loaded
            dataset = ''

        smooth_type = (f"{self.mode_selected.lower()}-smooth" if self.config == "cubeviz"
                       else "smooth")
        stddev = f"stddev-{self.stddev}"

        # Overriding is allowed, so do not check for uniqueness
        self.results_label_default = (
            self.app.return_data_label(f"{dataset}{smooth_type} {stddev}", check_unique=False))

    @observe("dataset_selected")
    def _update_viewer_filters(self, event={}):
        if not hasattr(self, 'dataset'):
            # during initial init, this can trigger before the component is initialized
            return

        if self.dataset.selected_dc_item is not None:
            selected_data_is_1d = len(self.dataset.selected_dc_item.data.shape) == 1

            if selected_data_is_1d:
                # only want spectral viewers in the options
                self.add_results.viewer.filters = ['is_spectrum_viewer']
            else:
                # only want image viewers in the options
                self.add_results.viewer.filters = ['is_image_viewer']

    @observe("mode_selected")
    def _update_dataset_viewer_filters(self, event={}):
        if event.get('new', self.mode_selected) == 'Spatial':
            # MUST be a cube
            self.dataset.add_filter('layer_in_flux_viewer')
        else:
            # can either be a cube OR a spectrum
            self.dataset.remove_filter('layer_in_flux_viewer')
        self._update_viewer_filters()

    def vue_apply(self, event={}):
        self.smooth(add_data=True)

    def smooth(self, add_data=True):
        """
        Smooth according to the settings in the plugin.

        Parameters
        ----------
        add_data : bool
            Whether to add the resulting trace to the application, according to the options
            defined in the plugin.

        Returns
        -------
        spec : `~specutils.Spectrum1D`
            The smoothed spectrum or data cube
        """
        if self.mode_selected == 'Spatial':
            if self.config != 'cubeviz':
                raise NotImplementedError("spatial smoothing only supported for Cubeviz")
            # TODO: in vuetify >2.3, timeout should be set to -1 to keep open
            #  indefinitely
            snackbar_message = SnackbarMessage(
                "Smoothing spatial slices of cube...",
                loading=True, timeout=0, sender=self)
            self.hub.broadcast(snackbar_message)

            results = self.spatial_smooth()

        else:
            results = self.spectral_smooth()

        if add_data:
            # add data to the collection/viewer
            self.add_results.add_results_from_plugin(results)
            self._set_default_results_label()

        snackbar_message = SnackbarMessage(
            f"Data set '{self.dataset_selected}' smoothed successfully.",
            color="success",
            sender=self)
        self.hub.broadcast(snackbar_message)

        return results

    @with_spinner()
    def spectral_smooth(self):
        """
        Smooth the input spectrum along the spectral axis.  To add the resulting spectrum into
        the app, set label options and use :meth:`smooth` instead.

        Returns
        -------
        spec : `~specutils.Spectrum1D`
            The smoothed spectrum
        """
        # Testing inputs to make sure putting smoothed spectrum into
        # datacollection works
        # input_flux = Quantity(np.array([0.2, 0.3, 2.2, 0.3]), u.Jy)
        # input_spaxis = Quantity(np.array([1, 2, 3, 4]), u.micron)
        # spec1 = Spectrum1D(input_flux, spectral_axis=input_spaxis)

        # Takes the user input from the dialog (stddev) and uses it to
        # define a standard deviation for gaussian smoothing
        cube = self.dataset.get_object(cls=Spectrum1D, statistic=None)
        spec_smoothed = gaussian_smooth(cube, stddev=self.stddev)

        return spec_smoothed

    @with_spinner('spinner')
    def spatial_smooth(self):
        """
        Use astropy convolution machinery to smooth the spatial dimensions of
        the data cube.  To add the resulting cube into
        the app, set label options and use :meth:`smooth` instead.

        Returns
        -------
        cube : `~specutils.Spectrum1D`
            The smoothed cube
        """
        cube = self.dataset.selected_obj
        flux_unit = cube.flux.unit

        # Extend the 2D kernel to have a length 1 spectral dimension, so that
        # we can do "3d" convolution to the whole cube
        kernel = np.expand_dims(Gaussian2DKernel(self.stddev), 2)

        convolved_data = convolve(cube, kernel)

        # Copy 3D WCS from input cube.
        data = self.dataset.selected_dc_item
        # Similar to coords_info logic.
        if '_orig_spec' in getattr(data, 'meta', {}):
            w = data.meta['_orig_spec'].wcs
        else:
            w = data.coords

        # Create a new cube with the old metadata. Note that astropy
        # convolution generates values for masked (NaN) data.
        newcube = Spectrum1D(flux=convolved_data * flux_unit, wcs=w)

        return newcube



import warnings
from pathlib import Path

import numpy as np
import pytest
from astropy import units as u
from astropy.nddata import CCDData
from astropy.tests.helper import assert_quantity_allclose
from astropy.wcs import WCS
from numpy.testing import assert_allclose
from specutils import SpectralRegion

from jdaviz.core.custom_units_and_equivs import PIX2, SPEC_PHOTON_FLUX_DENSITY_UNITS


@pytest.mark.parametrize("cube_type", ["Surface Brightness", "Flux"])
def test_user_api(cubeviz_helper, spectrum1d_cube, spectrum1d_cube_sb_unit, cube_type):

    # test is parameterize to test a cube that is in Jy / sr (Surface Brightness)
    # as well as Jy (Flux), to test that flux cubes, which are converted in the
    # parser to flux / pix^2 surface brightness cubes, both work correctly.

    if cube_type == 'Surface Brightness':
        cube = spectrum1d_cube_sb_unit
    elif cube_type == 'Flux':
        cube = spectrum1d_cube

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="No observer defined on WCS.*")
        cubeviz_helper.load_data(cube, data_label='test')

    mm = cubeviz_helper.plugins['Moment Maps']
    assert not mm._obj.continuum_marks['center'].visible
    with mm.as_active():
        assert mm._obj.continuum_marks['center'].visible
        mm.n_moment = 0
        # no continuum so marks should be empty
        assert len(mm._obj.continuum_marks['center'].x) == 0

        mom = mm.calculate_moment()

        mm.continuum = 'Surrounding'
        mm.continuum_width = 10
        assert len(mm._obj.continuum_marks['center'].x) > 0

        mom_sub = mm.calculate_moment()
        assert isinstance(mom_sub.wcs, WCS)

        assert mom != mom_sub

        mm.n_moment = -1
        with pytest.raises(ValueError, match="Moment must be a positive integer"):
            mm.calculate_moment()

        mm.n_moment = 1
        with pytest.raises(ValueError, match="not one of"):
            mm._obj.output_unit_selected = "Bad Unit"
        mm._obj.output_unit_selected = "Surface Brightness"
        with pytest.raises(ValueError, match="units must be in"):
            mm.calculate_moment()


@pytest.mark.parametrize("cube_type", ["Surface Brightness", "Flux"])
def test_moment_calculation(cubeviz_helper, spectrum1d_cube,
                            spectrum1d_cube_sb_unit, cube_type, tmp_path):

    moment_unit = "Jy m"
    moment_value_str = "+6.40166e-10"

    if cube_type == 'Surface Brightness':
        moment_unit += " / sr"
        cube = spectrum1d_cube_sb_unit
        cube_unit = cube.unit.to_string()

    elif cube_type == 'Flux':
        moment_unit += " / pix2"
        cube = spectrum1d_cube
        cube_unit = cube.unit.to_string() + " / pix2"  # cube in Jy will become cube in Jy / pix2

    dc = cubeviz_helper.app.data_collection
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="No observer defined on WCS.*")
        cubeviz_helper.load_data(cube, data_label='test')

    flux_viewer = cubeviz_helper.app.get_viewer(cubeviz_helper._default_flux_viewer_reference_name)

    # Since we are not really displaying, need this to trigger GUI stuff.
    flux_viewer.shape = (100, 100)
    flux_viewer.state._set_axes_aspect_ratio(1)

    mm = cubeviz_helper.plugins["Moment Maps"]
    mm.dataset = 'test[FLUX]'

    mm.n_moment = 0  # Collapsed sum, will get back 2D spatial image
    assert mm._obj.results_label == 'moment 0'

    # result is always a SB unit - if flux cube loaded, per pix2
    assert mm.output_unit == 'Surface Brightness'

    mm._obj.add_results.viewer.selected = cubeviz_helper._default_uncert_viewer_reference_name
    mm._obj.vue_calculate_moment()

    assert mm._obj.moment_available
    assert dc[-1].label == 'moment 0'
    mv_data = cubeviz_helper.app.get_viewer(
        cubeviz_helper._default_uncert_viewer_reference_name
    ).data()
    # by default, will overwrite the previous entry (so only one data entry)
    assert len(mv_data) == 1
    assert mv_data[0].label == 'moment 0'

    assert len(dc.links) == 19

    # label should remain unchanged but raise overwrite warnings
    assert mm._obj.results_label == 'moment 0'
    assert mm._obj.results_label_overwrite is True

    # Make sure coordinate display works in flux viewer (loaded data, not the moment map)
    label_mouseover = cubeviz_helper.app.session.application._tools['g-coords-info']
    label_mouseover._viewer_mouse_event(flux_viewer, {'event': 'mousemove',
                                                      'domain': {'x': 0, 'y': 0}})
    assert flux_viewer.state.slices == (0, 0, 1)
    # Slice 0 has 8 pixels, this is Slice 1
    assert label_mouseover.as_text() == (f"Pixel x=00.0 y=00.0 Value +8.00000e+00 {cube_unit}",
                                         "World 13h39m59.9731s +27d00m00.3600s (ICRS)",
                                         "204.9998877673 27.0001000000 (deg)")

    # Make sure adding it to viewer does not crash.
    cubeviz_helper.app.add_data_to_viewer(
        cubeviz_helper._default_flux_viewer_reference_name, 'moment 0'
    )

    result = dc[-1].get_object(cls=CCDData)
    assert result.shape == (2, 4)  # Cube shape is (2, 2, 4), moment transposes
    assert isinstance(dc[-1].coords, WCS)

    # Make sure coordinate display now show moment map info (no WCS)
    label_mouseover._viewer_mouse_event(flux_viewer, {'event': 'mousemove',
                                                      'domain': {'x': 0, 'y': 0}})

    # Slice 0 has 8 pixels, this is Slice 1
    assert label_mouseover.as_text() == (
        f"Pixel x=00.0 y=00.0 Value {moment_value_str} {moment_unit}",
        "World 13h39m59.9731s +27d00m00.3600s (ICRS)",
        "204.9998877673 27.0001000000 (deg)")

    mm._obj.n_moment = 1
    mm._obj.output_unit_selected = "Spectral Unit"
    assert mm._obj.results_label == 'moment 1'
    assert mm._obj.results_label_overwrite is False
    mm._obj.vue_calculate_moment()

    assert dc[-1].label == 'moment 1'

    assert len(dc.links) == 27
    assert len(dc.external_links) == 6  # pixel linked
    # Link 3D z to 2D x and 3D y to 2D y
    assert (dc.external_links[2].cids1[0].label == "Pixel Axis 0 [z]" and
            dc.external_links[2].cids2[0].label == "Pixel Axis 1 [x]" and
            dc.external_links[3].cids1[0].label == "Pixel Axis 1 [y]" and
            dc.external_links[3].cids2[0].label == "Pixel Axis 0 [y]")

    # Coordinate display should be unaffected.
    label_mouseover._viewer_mouse_event(flux_viewer, {'event': 'mousemove',
                                                      'domain': {'x': 0, 'y': 0}})
    assert label_mouseover.as_text() == (
        f"Pixel x=00.0 y=00.0 Value {moment_value_str} {moment_unit}",
        "World 13h39m59.9731s +27d00m00.3600s (ICRS)",
        "204.9998877673 27.0001000000 (deg)")


def test_moment_velocity_calculation(cubeviz_helper, spectrum1d_cube):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="No observer defined on WCS.*")
        cubeviz_helper.load_data(spectrum1d_cube, data_label='test')

    uncert_viewer = cubeviz_helper.app.get_viewer("uncert-viewer")

    # Since we are not really displaying, need this to trigger GUI stuff.
    uncert_viewer.shape = (100, 100)
    uncert_viewer.state._set_axes_aspect_ratio(1)

    mm = cubeviz_helper.plugins["Moment Maps"]
    mm._obj.dataset_selected = 'test[FLUX]'

    # Test moment 1 in velocity
    mm.n_moment = 1
    mm.add_results.viewer = "uncert-viewer"
    assert mm._obj.results_label == 'moment 1'
    mm.output_unit = "Velocity"

    with pytest.raises(ValueError, match="reference_wavelength must be set"):
        mm.calculate_moment()

    mm.reference_wavelength = 4.63e-7
    mm.calculate_moment()

    # Make sure coordinate display works
    label_mouseover = cubeviz_helper.app.session.application._tools['g-coords-info']
    label_mouseover._viewer_mouse_event(uncert_viewer, {'event': 'mousemove',
                                                        'domain': {'x': 0, 'y': 0}})
    assert label_mouseover.as_text() == ("Pixel x=00.0 y=00.0 Value -4.14668e+02 km / s",
                                         "World 13h39m59.9731s +27d00m00.3600s (ICRS)",
                                         "204.9998877673 27.0001000000 (deg)")

    # Add test for unit conversion
    assert mm._obj.output_radio_items[0]['unit_str'] == 'm'
    uc_plugin = cubeviz_helper.plugins['Unit Conversion']._obj
    uc_plugin.spectral_unit.selected = 'Angstrom'
    assert mm._obj.output_radio_items[0]['unit_str'] == 'Angstrom'
    uc_plugin.spectral_unit.selected = 'm'

    # Test moment 2 in velocity
    mm.n_moment = 2
    mm.calculate_moment()

    label_mouseover._viewer_mouse_event(uncert_viewer, {'event': 'mousemove',
                                                        'domain': {'x': 1, 'y': 1}})

    assert label_mouseover.as_text() == ("Pixel x=01.0 y=01.0 Value +2.32415e+01 km / s",
                                         "World 13h39m59.9461s +27d00m00.7200s (ICRS)",
                                         "204.9997755344 27.0001999998 (deg)")


def test_moment_frequency_unit_conversion(cubeviz_helper, spectrum1d_cube_larger):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="No observer defined on WCS.*")
        cubeviz_helper.load_data(spectrum1d_cube_larger, data_label='test')

    uc = cubeviz_helper.plugins['Unit Conversion']
    mm = cubeviz_helper.plugins['Moment Maps']

    unit = u.Unit(uc.spectral_unit.selected)
    cubeviz_helper.plugins['Subset Tools'].import_region(
        SpectralRegion(4.624e-07 * unit, 4.627e-07 * unit))

    uc.spectral_unit = 'Hz'
    mm.spectral_subset = 'Subset 1'
    mm.continuum = 'Surrounding'
    mm.n_moment = 1
    mm.output_unit = 'Spectral Unit'
    moment_1_data = mm.calculate_moment()
    mm.n_moment = 0
    moment_0_data = mm.calculate_moment()

    # Check to make sure there are no nans
    assert len(np.where(moment_1_data.data > 0)[0]) == 8
    assert_quantity_allclose(moment_0_data, -2.9607526e+09*u.Unit("Hz Jy / pix2"))


def test_write_momentmap(cubeviz_helper, spectrum1d_cube, tmp_path):
    ''' Test writing a moment map out to a FITS file on disk '''

    # Simulate an existing file on disk to check for overwrite warning
    test_file = tmp_path / "test_file.fits"
    existing_sentinel_text = "This is a simulated, existing file on disk"
    test_file.write_text(existing_sentinel_text)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="No observer defined on WCS.*")
        cubeviz_helper.load_data(spectrum1d_cube, data_label='test')
    plugin = cubeviz_helper.plugins['Moment Maps']
    plugin.calculate_moment()

    # and shouldn't have written anything (the file should be intact)
    assert test_file.read_text() == existing_sentinel_text

    label = plugin._obj.add_results.label
    export_plugin = cubeviz_helper.plugins['Export']._obj

    assert label in export_plugin.data_collection.labels


@pytest.mark.remote_data
def test_momentmap_nirspec_prism(cubeviz_helper, tmp_path):
    uri = "mast:jwst/product/jw02732-o003_t002_nirspec_prism-clear_s3d.fits"
    local_path = str(tmp_path / Path(uri).name)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        cubeviz_helper.load_data(uri, cache=True, local_path=local_path)
    uc = cubeviz_helper.plugins["Unit Conversion"]
    uc.open_in_tray()  # plugin has to be open for unit change to take hold
    uc._obj.show_translator = True
    uc.spectral_y_type.selected = 'Surface Brightness'
    mm = cubeviz_helper.plugins['Moment Maps']._obj
    mm.open_in_tray()  # plugin has to be open for unit change to take hold
    mm._set_data_units()
    mm.calculate_moment()
    assert isinstance(mm.moment.wcs, WCS)

    # Because cube axes order is re-arranged by specutils on load, this gets confusing.
    # There is a swapaxes within Moment Map WCS calculation.
    sky_moment = mm.moment.wcs.pixel_to_world(50, 30)
    sky_cube = cubeviz_helper.app.data_collection[0].meta["_orig_spec"].wcs.celestial.pixel_to_world(30, 50)  # noqa: E501
    assert_allclose((sky_moment.ra.deg, sky_moment.dec.deg),
                    (sky_cube.ra.deg, sky_cube.dec.deg))


def test_correct_output_spectral_y_units(cubeviz_helper, spectrum1d_cube_custom_fluxunit):

    moment_unit = "Jy m / sr"

    # test that the output unit labels in the moment map plugin reflect any
    # changes made in the unit conversion plugin.
    # NOTE: for now this is limited to moment 0, test should be expanded to
    # test higher-order moments once implemented.

    sb_cube = spectrum1d_cube_custom_fluxunit(fluxunit=u.MJy / u.sr)

    # load surface brigtness cube
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="No observer defined on WCS.*")
        cubeviz_helper.load_data(sb_cube, data_label='test')

    uc = cubeviz_helper.plugins["Unit Conversion"]
    uc.open_in_tray()  # plugin has to be open for unit change to take hold
    uc._obj.show_translator = True
    uc.spectral_y_type.selected = 'Surface Brightness'
    mm = cubeviz_helper.plugins['Moment Maps']._obj
    mm.open_in_tray()  # plugin has to be open for unit change to take hold
    mm._set_data_units()

    # check that label is initialized with 'Surface Brightness' since the cube
    # loaded is in MJy / sr. for the 0th moment, the only item will be the 0th
    # in this list. also check that the initial unit is correct.
    output_unit_moment_0 = mm.output_unit_items[0]
    assert output_unit_moment_0['label'] == 'Surface Brightness'
    assert output_unit_moment_0['unit_str'] == 'MJy m / sr'

    # check that calculated moment has the correct units of MJy m / sr
    mm.calculate_moment()
    assert mm.moment.unit == f'M{moment_unit}'

    # now change surface brightness units in the unit conversion plugin

    uc.flux_unit = 'Jy'

    # and make sure this change is propogated
    output_unit_moment_0 = mm.output_unit_items[0]
    assert output_unit_moment_0['label'] == 'Surface Brightness'
    assert output_unit_moment_0['unit_str'] == 'Jy m / sr'

    # and that calculated moment has the correct units
    mm.calculate_moment()
    assert mm.moment.unit == moment_unit

    # now change the spectral unit and make sure that change is
    # reflected in MM plugin
    uc.spectral_unit = 'um'
    assert output_unit_moment_0['unit_str'] == 'Jy um / sr'

    mm.calculate_moment()
    assert mm.moment.unit == moment_unit.replace('m', 'um')


@pytest.mark.parametrize("flux_unit", [u.Unit(x) for x in SPEC_PHOTON_FLUX_DENSITY_UNITS][1:2])
@pytest.mark.parametrize("angle_unit", [u.sr, PIX2])
@pytest.mark.parametrize("new_flux_unit", [u.Unit(x) for x in SPEC_PHOTON_FLUX_DENSITY_UNITS][1:2])
def test_moment_zero_unit_flux_conversions(cubeviz_helper,
                                           spectrum1d_cube_custom_fluxunit,
                                           flux_unit, angle_unit, new_flux_unit):
    """
    Test cubeviz moment maps with all possible unit conversions for
    cubes in spectral/photon surface brightness units (e.g. Jy/sr, Jy/pix2).

    The moment map plugin should respect the choice of flux and angle
    unit selected in the Unit Conversion plugin, and inputs and results should
    be converted based on selection. All conversions between units in the
    flux dropdown menu in the unit conversion plugin should be supported
    by moment maps.
    """

    if new_flux_unit == flux_unit:  # skip 'converting' to same unit
        return
    new_flux_unit_str = new_flux_unit.to_string()

    cube_unit = flux_unit / angle_unit

    sb_cube = spectrum1d_cube_custom_fluxunit(fluxunit=cube_unit)

    # load surface brigtness cube
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="No observer defined on WCS.*")
        cubeviz_helper.load_data(sb_cube, data_label='test')

    # get plugins
    uc = cubeviz_helper.plugins["Unit Conversion"]
    mm = cubeviz_helper.plugins['Moment Maps']._obj

    # and flux viewer for mouseover info
    flux_viewer = cubeviz_helper.app.get_viewer(cubeviz_helper._default_flux_viewer_reference_name)
    label_mouseover = cubeviz_helper.app.session.application._tools['g-coords-info']

    # convert to new flux unit
    uc.flux_unit.selected = new_flux_unit_str

    new_mm_unit = (new_flux_unit * u.m / u.Unit(angle_unit)).to_string()
    assert mm.output_unit_items[0]['label'] == 'Surface Brightness'
    assert mm.output_unit_items[0]['unit_str'] == new_mm_unit

    # calculate moment with new output label and plot in flux viewer
    mm.add_results.label = new_flux_unit_str
    mm.add_results.viewer.selected = cubeviz_helper._default_flux_viewer_reference_name
    mm.calculate_moment()

    assert mm.moment.unit == new_mm_unit

    # make sure mouseover info in flux unit is new moment map unit
    # which should be flux/sb unit times spectral axis unit (e.g. MJy m / sr)
    label_mouseover._viewer_mouse_event(flux_viewer,
                                        {'event': 'mousemove',
                                         'domain': {'x': 0, 'y': 0}})
    m_orig = label_mouseover.as_text()[0]
    assert ((new_flux_unit / angle_unit) * u.m).to_string() in m_orig

    # 'jiggle' mouse so we can move it back and compare original coordinate
    label_mouseover._viewer_mouse_event(flux_viewer,
                                        {'event': 'mousemove',
                                         'domain': {'x': 1, 'y': 1}})

    # when flux unit is changed, the mouseover unit conversion should be
    # skipped so that the plotted moment map remains in its original
    # unit. setting back to the original flux unit also ensures that
    # each iteration begins on the same unit so that every comparison
    # is tested
    uc.flux_unit.selected = new_flux_unit_str
    label_mouseover._viewer_mouse_event(flux_viewer,
                                        {'event': 'mousemove',
                                         'domain': {'x': 0, 'y': 0}})
    assert m_orig == label_mouseover.as_text()[0]



import astropy.units as u

__all__ = ["PIX2", "SPEC_PHOTON_FLUX_DENSITY_UNITS",
           "_eqv_pixar_sr", "_eqv_flux_to_sb_pixel",
           "_eqv_sb_per_pixel_to_per_angle"]

# define custom composite units here
PIX2 = u.pix * u.pix


# Add spaxel to enabled units
def enable_spaxel_unit():
    spaxel = u.Unit('spaxel', represents=u.pixel, parse_strict='silent')
    u.add_enabled_units([spaxel])
    return


def _spectral_and_photon_flux_density_units(freq_only=False, wav_only=False,
                                            as_units=False):
    """
    Returns an alphabetically sorted list of string representations
    of flux density units supported by the unit conversion plugin. These units can
    be used for conversion if the input data units are compatible (i.e., directly
    equivalent or convertible using ``u.spectral_density(cube_wave)``). If both
    ``freq_only`` and ``wav_only`` are False, the function returns the combined
    list of all supported units. If ``as_units`` is True, return ``u.Unit`` instead
    of strings.

    Parameters:
    -----------
    freq_only : bool, optional
        If True, returns only frequency-based flux density units. (Default=False).

    wav_only : bool, optional
        If True, returns only wavelength-based flux density units. (Default=False).

    as_units : bool, optional
        If True, return ``u.Unit`` instead of strings

    Returns:
    --------
    list of str
        A sorted list of flux density unit strings.
    """
    wav_flux_units = ['erg / (Angstrom s cm2)', 'ph / (Angstrom s cm2)']
    freq_flux_units = ['Jy', 'mJy', 'uJy', 'MJy', 'W / (Hz m2)', 'eV / (Hz s m2)',
                       'erg / (Hz s cm2)', 'ph / (Hz s cm2)']

    if freq_only:
        flux_units = sorted(freq_flux_units)
    elif wav_only:
        flux_units = sorted(wav_flux_units)
    else:
        flux_units = sorted(freq_flux_units + wav_flux_units)

    if as_units:
        flux_units = [u.Unit(x) for x in flux_units]

    return flux_units


SPEC_PHOTON_FLUX_DENSITY_UNITS = _spectral_and_photon_flux_density_units()


def _eqv_pixar_sr(pixar_sr):
    """
    Return Equivalencies to convert from flux to flux per solid
    angle (aka surface brightness) using scale ratio ``pixar_sr``
    (steradians per pixel).
    """
    def converter_flux(x):  # Surface Brightness -> Flux
        return x * pixar_sr

    def iconverter_flux(x):  # Flux -> Surface Brightness
        return x / pixar_sr

    return [
        (u.MJy / u.sr, u.MJy, converter_flux, iconverter_flux),
        (u.erg / (u.s * u.cm**2 * u.Angstrom * u.sr), u.erg / (u.s * u.cm**2 * u.Angstrom), converter_flux, iconverter_flux),  # noqa
        (u.ph / (u.Angstrom * u.s * u.cm**2 * u.sr), u.ph / (u.Angstrom * u.s * u.cm**2), converter_flux, iconverter_flux),  # noqa
        (u.ph / (u.Hz * u.s * u.cm**2  * u.sr), u.ph / (u.Hz * u.s * u.cm**2), converter_flux, iconverter_flux),  # noqa
        (u.ct / u.sr, u.ct, converter_flux, iconverter_flux)  # noqa
    ]


def _eqv_flux_to_sb_pixel():
    """
    Returns an Equivalency between ``flux_unit`` and ``flux_unit`/pix**2``. This
    allows conversion between flux and flux-per-square-pixel surface brightness
    e.g MJy <> MJy / pix2
    """

    # generate an equivalency for each flux type that would need
    # another equivalency for converting to/from
    flux_units = [u.MJy,
                  u.erg / (u.s * u.cm**2 * u.Angstrom),
                  u.ph / (u.Angstrom * u.s * u.cm**2),
                  u.ph / (u.Hz * u.s * u.cm**2),
                  u.ct,
                  u.DN,
                  u.DN / u.s]

    equivs = [(flux_unit, flux_unit / PIX2, lambda x: x, lambda x: x)
              for flux_unit in flux_units]

    # We also need to convert between spaxel and pixel squared
    equivs += [(flux_unit / u.Unit('spaxel'), flux_unit / PIX2,
               lambda x: x, lambda x: x) for flux_unit in flux_units]

    return equivs


def _eqv_sb_per_pixel_to_per_angle(flux_unit, scale_factor=1):
    """
    Returns an equivalency between ``flux_unit`` per square pixel and
    ``flux_unit`` per solid angle to be able to compare and convert between units
    like Jy/pix**2 and Jy/sr. The scale factor is assumed to be in steradians,
    to follow the convention of the PIXAR_SR keyword.
    Note:
    To allow conversions between units like ``ph / (Hz s cm2 sr)`` and
    MJy / pix2, which would require this equivalency as well as u.spectral_density,
    these CAN'T be combined when converting like:
    equivalencies=u.spectral_density(1 * u.m) + _eqv_sb_per_pixel_to_per_angle(u.Jy)
    So additional logic is needed to compare units that need both equivalencies
    (one solution being creating this equivalency for each equivalent flux-type.)

    """

    # the two types of units we want to define a conversion between
    flux_solid_ang = flux_unit / u.sr
    flux_sq_pix = flux_unit / PIX2

    pix_to_solid_angle_equiv = [(flux_solid_ang, flux_sq_pix,
                                lambda x: x * scale_factor,
                                lambda x: x / scale_factor)]

    return pix_to_solid_angle_equiv



from zipfile import ZipFile

import numpy as np
import pytest
from astropy import units as u
from astropy.io import fits
from astropy.tests.helper import assert_quantity_allclose
from specutils import Spectrum1D, SpectrumList, SpectrumCollection, SpectralRegion
from astropy.utils.data import download_file

from jdaviz.app import Application
from jdaviz.core.marks import LineUncertainties
from jdaviz import Specviz


class TestSpecvizHelper:
    @pytest.fixture(autouse=True)
    def setup_class(self, specviz_helper, spectrum1d, multi_order_spectrum_list):
        self.spec_app = specviz_helper
        self.spec = spectrum1d
        self.spec_list = SpectrumList([spectrum1d] * 3)
        self.multi_order_spectrum_list = multi_order_spectrum_list

        self.label = "Test 1D Spectrum"
        self.spec_app.load_data(spectrum1d, data_label=self.label)

    def test_load_spectrum1d(self):
        # starts with a single loaded spectrum1d object:
        assert len(self.spec_app.app.data_collection) == 1
        dc_0 = self.spec_app.app.data_collection[0]
        assert dc_0.label == self.label
        assert dc_0.meta['uncertainty_type'] == 'std'

        data = self.spec_app.get_data()

        assert isinstance(data, Spectrum1D)

    def test_load_hdulist(self):
        # Create a fake fits file with a 1D spectrum for testing.
        primary_header = fits.Header({'TELESCOP': 'Fake Telescope'})
        primary_hdu = fits.PrimaryHDU(header=primary_header)

        # Image extension HDU with a 1D spectrum
        wavelength = np.linspace(5000, 6000, 100)
        flux = np.ones(100)
        spectrum_table = fits.BinTableHDU.from_columns([
            fits.Column(name='WAVELENGTH', array=wavelength, format='E', unit="Angstrom"),
            fits.Column(name='FLUX', array=flux, format='E', unit="Jy")
        ])
        spectrum_table.header['INSTRUME'] = 'Fake Instrument'
        fake_hdulist = fits.HDUList([primary_hdu, spectrum_table])
        self.label = "Test 1D Spectrum"
        self.spec_app.load_data(fake_hdulist)
        data = self.spec_app.get_data(data_label=self.label)
        # HDUList should load as Spectrum1D
        assert isinstance(data, Spectrum1D)

    def test_load_spectrum_list_no_labels(self):
        # now load three more spectra from a SpectrumList, without labels
        self.spec_app.load_data(self.spec_list)
        assert len(self.spec_app.app.data_collection) == 4
        for i in (1, 2, 3):
            assert "specviz_data" in self.spec_app.app.data_collection[i].label

    def test_load_spectrum_list_with_labels(self):
        # now load three more spectra from a SpectrumList, with labels:
        labels = ["List test 1", "List test 2", "List test 3"]
        self.spec_app.load_data(self.spec_list, data_label=labels)
        assert len(self.spec_app.app.data_collection) == 4

    def test_load_multi_order_spectrum_list(self):
        assert len(self.spec_app.app.data_collection) == 1
        # now load ten spectral orders from a SpectrumList:
        self.spec_app.load_data(self.multi_order_spectrum_list)
        assert len(self.spec_app.app.data_collection) == 11

    def test_mismatched_label_length(self):
        with pytest.raises(ValueError, match='Length'):
            labels = ["List test 1", "List test 2"]
            self.spec_app.load_data(self.spec_list, data_label=labels)

    def test_load_spectrum_collection(self):
        with pytest.raises(TypeError):
            collection = SpectrumCollection([1]*u.AA)
            self.spec_app.load_data(collection)

    def test_get_spectra(self):
        with pytest.warns(UserWarning, match='Applying the value from the redshift slider'):
            spectra = self.spec_app.get_spectra()

        assert_quantity_allclose(spectra[self.label].flux,
                                 self.spec.flux, atol=1e-5*u.Unit(self.spec.flux.unit))

    def test_get_spectra_no_redshift(self):
        spectra = self.spec_app.get_spectra(apply_slider_redshift=None)

        assert_quantity_allclose(spectra[self.label].flux,
                                 self.spec.flux, atol=1e-5*u.Unit(self.spec.flux.unit))

    def test_get_spectra_no_data_label(self):
        spectra = self.spec_app.get_spectra(data_label=None, apply_slider_redshift=True)

        assert_quantity_allclose(spectra[self.label].flux,
                                 self.spec.flux, atol=1e-5*u.Unit(self.spec.flux.unit))

    def test_get_spectra_label_redshift(self):
        spectra = self.spec_app.get_spectra(data_label=self.label, apply_slider_redshift=True)

        assert_quantity_allclose(spectra.flux,
                                 self.spec.flux, atol=1e-5*u.Unit(self.spec.flux.unit))

    def test_get_spectra_label_redshift_warn(self):
        with pytest.warns(UserWarning, match='Applying the value from the redshift slider'):
            spectra = self.spec_app.get_spectra(data_label=self.label, apply_slider_redshift="Warn")

        assert_quantity_allclose(spectra.flux,
                                 self.spec.flux, atol=1e-5*u.Unit(self.spec.flux.unit))

    def test_get_spectral_regions_none(self):
        plg = self.spec_app.plugins['Subset Tools']
        spec_region = plg.get_regions()

        assert spec_region == {}

    def test_get_spectral_regions_one(self):
        self.spec_app.plugins['Subset Tools'].import_region(
            SpectralRegion(6000*self.spec.spectral_axis.unit, 6500*self.spec.spectral_axis.unit))
        plg = self.spec_app.plugins['Subset Tools']
        spec_region = plg.get_regions()
        assert len(spec_region['Subset 1'].subregions) == 1

    def test_get_spectral_regions_two(self):
        subset = (SpectralRegion(6000*self.spec.spectral_axis.unit,
                                 6500*self.spec.spectral_axis.unit) +
                  SpectralRegion(7300*self.spec.spectral_axis.unit,
                                 7800*self.spec.spectral_axis.unit))
        self.spec_app.plugins['Subset Tools'].import_region(subset, combination_mode='or')

        spec_region = self.spec_app.plugins['Subset Tools'].get_regions()

        assert len(spec_region['Subset 1'].subregions) == 2

    def test_get_spectral_regions_three(self):
        subset = (SpectralRegion(6000*self.spec.spectral_axis.unit,
                                 6400*self.spec.spectral_axis.unit) +
                  SpectralRegion(6600*self.spec.spectral_axis.unit,
                                 7000*self.spec.spectral_axis.unit) +
                  SpectralRegion(7300*self.spec.spectral_axis.unit,
                                 7800*self.spec.spectral_axis.unit))
        self.spec_app.plugins['Subset Tools'].import_region(subset, combination_mode='or')

        spec_region = self.spec_app.plugins['Subset Tools'].get_regions()

        assert len(spec_region['Subset 1'].subregions) == 3
        # Assert correct values for test with 3 subregions
        assert_quantity_allclose(spec_region['Subset 1'].subregions[0][0].value,
                                 6000., atol=1e-5)
        assert_quantity_allclose(spec_region['Subset 1'].subregions[0][1].value,
                                 6400., atol=1e-5)

        assert_quantity_allclose(spec_region['Subset 1'].subregions[1][0].value,
                                 6600., atol=1e-5)
        assert_quantity_allclose(spec_region['Subset 1'].subregions[1][1].value,
                                 7000., atol=1e-5)

        assert_quantity_allclose(spec_region['Subset 1'].subregions[2][0].value,
                                 7300., atol=1e-5)
        assert_quantity_allclose(spec_region['Subset 1'].subregions[2][1].value,
                                 7800., atol=1e-5)

    def test_get_spectral_regions_does_not_raise_value_error(self):
        subset = (SpectralRegion(1*self.spec.spectral_axis.unit,
                                 3*self.spec.spectral_axis.unit) +
                  SpectralRegion(4*self.spec.spectral_axis.unit,
                                 6*self.spec.spectral_axis.unit))
        self.spec_app.plugins['Subset Tools'].import_region(subset, combination_mode='or')

        spec_region = self.spec_app.plugins['Subset Tools'].get_regions()
        assert_quantity_allclose(spec_region['Subset 1'].subregions[0][0].value,
                                 1, atol=1e-5)
        assert_quantity_allclose(spec_region['Subset 1'].subregions[0][1].value,
                                 3, atol=1e-5)

        assert_quantity_allclose(spec_region['Subset 1'].subregions[1][0].value,
                                 4, atol=1e-5)
        assert_quantity_allclose(spec_region['Subset 1'].subregions[1][1].value,
                                 6, atol=1e-5)

    def test_get_spectral_regions_composite_region(self):
        subset = (SpectralRegion(6000*self.spec.spectral_axis.unit,
                                 7400*self.spec.spectral_axis.unit) +
                  SpectralRegion(6600*self.spec.spectral_axis.unit,
                                 7000*self.spec.spectral_axis.unit) +
                  SpectralRegion(7300*self.spec.spectral_axis.unit,
                                 7800*self.spec.spectral_axis.unit))
        self.spec_app.plugins['Subset Tools'].import_region(
            subset, combination_mode=['new', 'andnot', 'and'])

        spec_region = self.spec_app.plugins['Subset Tools'].get_regions()

        assert len(spec_region['Subset 1'].subregions) == 1
        # Assert correct values for test with 3 subregions
        assert_quantity_allclose(spec_region['Subset 1'].subregions[0][0].value,
                                 7300., atol=1e-5)
        assert_quantity_allclose(spec_region['Subset 1'].subregions[0][1].value,
                                 7400., atol=1e-5)

    def test_get_spectral_regions_composite_region_multiple_and_nots(self):
        subset = (SpectralRegion(6000*self.spec.spectral_axis.unit,
                                 7800*self.spec.spectral_axis.unit) +
                  SpectralRegion(6200*self.spec.spectral_axis.unit,
                                 6600*self.spec.spectral_axis.unit) +
                  SpectralRegion(7300*self.spec.spectral_axis.unit,
                                 7700*self.spec.spectral_axis.unit))
        self.spec_app.plugins['Subset Tools'].import_region(
            subset, combination_mode=['new', 'andnot', 'andnot'])

        spec_region = self.spec_app.plugins['Subset Tools'].get_regions()

        assert len(spec_region['Subset 1'].subregions) == 3
        # Assert correct values for test with 3 subregions
        assert_quantity_allclose(spec_region['Subset 1'].subregions[0][0].value,
                                 6000., atol=1e-5)
        assert_quantity_allclose(spec_region['Subset 1'].subregions[0][1].value,
                                 6200., atol=1e-5)

        assert_quantity_allclose(spec_region['Subset 1'].subregions[1][0].value,
                                 6600., atol=1e-5)
        assert_quantity_allclose(spec_region['Subset 1'].subregions[1][1].value,
                                 7300., atol=1e-5)
        assert_quantity_allclose(spec_region['Subset 1'].subregions[2][0].value,
                                 7700., atol=1e-5)
        assert_quantity_allclose(spec_region['Subset 1'].subregions[2][1].value,
                                 7800., atol=1e-5)


def test_get_spectra_no_spectra(specviz_helper, spectrum1d):
    with pytest.warns(UserWarning, match='Applying the value from the redshift slider'):
        spectra = specviz_helper.get_spectra()

    assert spectra == {}


def test_get_spectra_no_spectra_redshift_error(specviz_helper, spectrum1d):
    spectra = specviz_helper.get_spectra(apply_slider_redshift=True)

    assert spectra == {}


def test_get_spectra_no_spectra_label(specviz_helper, spectrum1d):
    label = "label"
    with pytest.raises(ValueError):
        specviz_helper.get_spectra(data_label=label)


def test_get_spectra_no_spectra_label_redshift_error(specviz_helper, spectrum1d):
    label = "label"
    with pytest.raises(ValueError):
        specviz_helper.get_spectra(data_label=label, apply_slider_redshift=True)


def test_add_spectrum_after_subset(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d, data_label="test")
    subset = SpectralRegion(6200 * spectrum1d.spectral_axis.unit,
                            7000 * spectrum1d.spectral_axis.unit)
    specviz_helper.plugins['Subset Tools'].import_region(subset)

    new_spec = specviz_helper.get_spectra(apply_slider_redshift=True)["test"]*0.9
    specviz_helper.load_data(new_spec, data_label="test2")


def test_get_spectral_regions_unit(specviz_helper, spectrum1d):
    # Ensure units we put in are the same as the units we get out
    specviz_helper.load_data(spectrum1d)
    subset = SpectralRegion(6200 * spectrum1d.spectral_axis.unit,
                            7000 * spectrum1d.spectral_axis.unit)
    specviz_helper.plugins['Subset Tools'].import_region(subset)

    subsets = specviz_helper.plugins['Subset Tools'].get_regions()
    reg = subsets.get('Subset 1')

    assert spectrum1d.spectral_axis.unit == reg.lower.unit
    assert spectrum1d.spectral_axis.unit == reg.upper.unit


def test_get_spectral_regions_unit_conversion(specviz_helper, spectrum1d):
    spec_viewer = specviz_helper.app.get_viewer('spectrum-viewer')

    # Mouseover without data should not crash.
    label_mouseover = specviz_helper.app.session.application._tools['g-coords-info']
    label_mouseover._viewer_mouse_event(spec_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 6100, 'y': 12.5}})
    assert label_mouseover.as_text() == ('', '', '')
    assert label_mouseover.icon == ''

    # If the reference (visible) data changes via unit conversion,
    # check that the region's units convert too
    specviz_helper.load_data(spectrum1d)  # Originally Angstrom

    # Also check coordinates info panel.
    # x=0 -> 6000 A, x=1 -> 6222.222 A
    label_mouseover._viewer_mouse_event(spec_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 6100, 'y': 12.5}})
    assert label_mouseover.as_text() == ('Cursor 6.10000e+03, 1.25000e+01',
                                         'Wave 6.00000e+03 Angstrom (0 pix)',  # actual 0.4
                                         'Flux 1.24967e+01 Jy')
    assert label_mouseover.icon == 'a'

    label_mouseover._viewer_mouse_event(spec_viewer,
                                        {'event': 'mousemove', 'domain': {'x': None, 'y': 12.5}})
    assert label_mouseover.as_text() == ('', '', '')
    assert label_mouseover.icon == ''

    # Convert the wavelength axis to micron
    new_spectral_axis = "um"
    specviz_helper.plugins['Unit Conversion'].spectral_unit = new_spectral_axis
    spectral_axis_unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    subset = SpectralRegion(0.6 * spectral_axis_unit, 0.7 * spectral_axis_unit)
    specviz_helper.plugins['Subset Tools'].import_region(subset)

    # Retrieve the Subset
    ss = specviz_helper.plugins['Subset Tools'].get_regions(use_display_units=False)
    reg = ss.get('Subset 1')
    assert reg.lower.unit == u.Angstrom
    assert reg.upper.unit == u.Angstrom

    ss = specviz_helper.plugins['Subset Tools'].get_regions(use_display_units=True)
    reg = ss.get('Subset 1')
    assert reg.lower.unit == u.um
    assert reg.upper.unit == u.um

    # Coordinates info panel should show new unit
    label_mouseover._viewer_mouse_event(spec_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 0.61, 'y': 12.5}})
    label_mouseover.as_text() == ('Cursor 6.10000e-01, 1.25000e+01',
                                  'Wave 6.00000e-01 micron (0 pix)',
                                  'Flux 1.24967e+01 Jy')
    assert label_mouseover.icon == 'a'

    label_mouseover._viewer_mouse_event(spec_viewer, {'event': 'mouseleave'})
    assert label_mouseover.as_text() == ('', '', '')
    assert label_mouseover.icon == ''


def test_subset_default_thickness(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)

    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    sv.toolbar.active_tool = sv.toolbar.tools['bqplot:xrange']

    spectral_axis_unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)
    subset = SpectralRegion(2.5 * spectral_axis_unit,
                            3.5 * spectral_axis_unit)
    specviz_helper.plugins['Subset Tools'].import_region(subset)
    # _on_layers_update is not triggered within CI
    sv._on_layers_update()
    assert sv.state.layers[-1].linewidth == 3


def test_app_links(specviz_helper, spectrum1d):
    sv = specviz_helper.app.get_viewer('spectrum-viewer')
    assert isinstance(sv.jdaviz_app, Application)
    assert isinstance(sv.jdaviz_helper, Specviz)


@pytest.mark.remote_data
def test_load_spectrum_list_directory(tmpdir, specviz_helper):
    # niriss_parser_test_data.zip
    test_data = 'https://stsci.box.com/shared/static/7ndets0vjjsa97la2hvjvm6xvnu5b594.zip'
    fn = download_file(test_data, cache=True, timeout=30)
    with ZipFile(fn, 'r') as sample_data_zip:
        sample_data_zip.extractall(tmpdir.strpath)
    data_path = str(tmpdir.join('NIRISS_for_parser_p0171'))

    # Load two NIRISS x1d files from FITS. They have 19 and 20 EXTRACT1D
    # extensions per file, for a total of 39 spectra to load:
    with pytest.warns(UserWarning, match='SRCTYPE is missing or UNKNOWN in JWST x1d loader'):
        specviz_helper.load_data(data_path)

    # NOTE: the length was 3 before specutils 1.9 (https://github.com/astropy/specutils/pull/982)
    expected_len = 39
    assert len(specviz_helper.app.data_collection) == expected_len

    for data in specviz_helper.app.data_collection:
        assert data.main_components[:2] == ['flux', 'uncertainty']

    dc_0 = specviz_helper.app.data_collection[0]
    assert 'header' not in dc_0.meta
    assert dc_0.meta['SPORDER'] == 1


@pytest.mark.remote_data
def test_load_spectrum_list_directory_concat(tmpdir, specviz_helper):
    # niriss_parser_test_data.zip
    test_data = 'https://stsci.box.com/shared/static/7ndets0vjjsa97la2hvjvm6xvnu5b594.zip'
    fn = download_file(test_data, cache=True, timeout=30)
    with ZipFile(fn, 'r') as sample_data_zip:
        sample_data_zip.extractall(tmpdir.strpath)
    data_path = str(tmpdir.join('NIRISS_for_parser_p0171'))

    # Load two x1d files from FITS. They have 19 and 20 EXTRACT1D
    # extensions per file, for a total of 39 spectra to load. Also concatenate
    # spectra common to each file into one "Combined" spectrum to load per file.
    # Now the total is (19 EXTRACT 1D + 1 Combined) + (20 EXTRACT 1D + 1 Combined) = 41.
    with pytest.warns(UserWarning, match='SRCTYPE is missing or UNKNOWN in JWST x1d loader'):
        specviz_helper.load_data(data_path, concat_by_file=True)
    assert len(specviz_helper.app.data_collection) == 41


def test_load_2d_flux(specviz_helper):
    # Test loading a spectrum with a 2D flux, which should be split into separate
    # 1D Spectrum1D objects to load in Specviz.
    spec = Spectrum1D(spectral_axis=np.linspace(4000, 6000, 10)*u.Angstrom,
                      flux=np.ones((4, 10))*u.Unit("1e-17 erg / (Angstrom cm2 s)"))
    specviz_helper.load_data(spec, data_label="test")

    assert len(specviz_helper.app.data_collection) == 4
    assert specviz_helper.app.data_collection[0].label == "test [0]"

    spec2 = Spectrum1D(spectral_axis=np.linspace(4000, 6000, 10)*u.Angstrom,
                       flux=np.ones((2, 10))*u.Unit("1e-17 erg / (Angstrom cm2 s)"))

    # Make sure 2D spectra in a SpectrumList also get split properly.
    spec_list = SpectrumList([spec, spec2])
    specviz_helper.load_data(spec_list, data_label="second test")

    assert len(specviz_helper.app.data_collection) == 10
    assert specviz_helper.app.data_collection[-1].label == "second test [5]"


def test_plot_uncertainties(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)

    specviz_viewer = specviz_helper.app.get_viewer("spectrum-viewer")

    assert len([m for m in specviz_viewer.figure.marks if isinstance(m, LineUncertainties)]) == 0

    specviz_viewer.state.show_uncertainty = True
    uncert_marks = [m for m in specviz_viewer.figure.marks if isinstance(m, LineUncertainties)]
    assert len(uncert_marks) == 1
    # mark has a lower and upper boundary, but each should have the same length as the spectrum
    assert len(uncert_marks[0].x[0]) == len(spectrum1d.flux)

    # enable as-steps and make sure doesn't crash (won't change the number of marks)
    specviz_viewer.state.layers[0].as_steps = True
    specviz_viewer._plot_uncertainties()
    uncert_marks = [m for m in specviz_viewer.figure.marks if isinstance(m, LineUncertainties)]
    assert len(uncert_marks) == 1
    # now the mark should have double the length as its drawn on the bin edges
    assert len(uncert_marks[0].x[0]) == len(spectrum1d.flux) * 2

    specviz_viewer.state.show_uncertainty = False

    assert len([m for m in specviz_viewer.figure.marks if isinstance(m, LineUncertainties)]) == 0


# Some API might be going through deprecation, so ignore the warning.
@pytest.mark.filterwarnings("ignore::DeprecationWarning")
def test_plugin_user_apis(specviz_helper):
    for plugin_name, plugin_api in specviz_helper.plugins.items():
        plugin = plugin_api._obj
        for attr in plugin_api._expose:
            assert hasattr(plugin, attr)


def test_data_label_as_posarg(specviz_helper, spectrum1d):
    # Passing in data_label keyword as posarg.
    specviz_helper.load_data(spectrum1d, 'my_spec')
    assert specviz_helper.app.data_collection[0].label == 'my_spec'


def test_spectra_partial_overlap(specviz_helper):
    spec_viewer = specviz_helper.app.get_viewer('spectrum-viewer')

    wave_1 = np.linspace(6000, 7000, 10) * u.AA
    flux_1 = ([1200] * wave_1.size) * u.nJy
    sp_1 = Spectrum1D(flux=flux_1, spectral_axis=wave_1)

    wave_2 = wave_1 + (800 * u.AA)
    flux_2 = ([60] * wave_2.size) * u.nJy
    sp_2 = Spectrum1D(flux=flux_2, spectral_axis=wave_2)

    specviz_helper.load_data(sp_1, data_label='left')
    specviz_helper.load_data(sp_2, data_label='right')

    # Test mouseover outside of left but in range for right.
    # Should show right spectrum even when mouse is near left flux.
    label_mouseover = specviz_helper.app.session.application._tools['g-coords-info']
    label_mouseover._viewer_mouse_event(spec_viewer,
                                        {'event': 'mousemove', 'domain': {'x': 7022, 'y': 1000}})
    assert label_mouseover.as_text() == ('Cursor 7.02200e+03, 1.00000e+03',
                                         'Wave 7.02222e+03 Angstrom (2 pix)',
                                         'Flux 6.00000e+01 nJy')
    assert label_mouseover.icon == 'b'


def test_spectra_incompatible_flux(specviz_helper):
    """https://github.com/spacetelescope/jdaviz/issues/2459"""
    wav = [1.1, 1.2, 1.3] * u.um
    sp1 = Spectrum1D(flux=[1, 1.1, 1] * (u.MJy / u.sr), spectral_axis=wav)
    sp2 = Spectrum1D(flux=[1, 1, 1.1] * (u.MJy), spectral_axis=wav)
    flux3 = ([1, 1.1, 1] * u.MJy).to(u.erg / u.s / u.cm / u.cm / u.AA, u.spectral_density(wav))
    sp3 = Spectrum1D(flux=flux3, spectral_axis=wav)

    specviz_helper.load_data(sp2, data_label="2")  # OK
    specviz_helper.load_data(sp1, data_label="1")  # Not OK
    specviz_helper.load_data(sp3, data_label="3")  # OK

    assert specviz_helper.app.data_collection.labels == ["2", "3"]


def test_delete_data_with_subsets(specviz_helper, spectrum1d, spectrum1d_nm):
    specviz_helper.load_data(spectrum1d, 'my_spec_AA')
    specviz_helper.load_data(spectrum1d_nm, 'my_spec_nm')

    spectral_axis_unit = u.Unit(specviz_helper.plugins['Unit Conversion'].spectral_unit.selected)

    subset = SpectralRegion(6200 * spectral_axis_unit,
                            7000 * spectral_axis_unit)
    specviz_helper.plugins['Subset Tools'].import_region(subset)

    assert len(specviz_helper.app.data_collection.subset_groups) == 1
    subset1 = specviz_helper.app.data_collection.subset_groups[0]
    assert subset1.subset_state.att.parent.label == "my_spec_AA"
    np.testing.assert_allclose((subset1.subset_state.lo, subset1.subset_state.hi), (6200, 7000))

    specviz_helper.app.remove_data_from_viewer("spectrum-viewer", "my_spec_AA")
    specviz_helper.app.data_item_remove("my_spec_AA")

    # Check that the reparenting and coordinate recalculations happened
    assert subset1.subset_state.att.parent.label == "my_spec_nm"
    np.testing.assert_allclose((subset1.subset_state.lo, subset1.subset_state.hi), (620, 700))



from echo import delay_callback

import numpy as np

from glue.config import data_translator
from glue.core import BaseData
from glue.core.exceptions import IncompatibleAttribute
from glue.core.units import UnitConverter
from glue.core.subset import Subset
from glue.core.subset_group import GroupedSubset
from glue.viewers.scatter.state import ScatterLayerState as BqplotScatterLayerState

from glue_astronomy.spectral_coordinates import SpectralCoordinates
from glue_jupyter.bqplot.profile import BqplotProfileView
from glue_jupyter.bqplot.image import BqplotImageView
from glue_jupyter.table import TableViewer

from astropy.utils import deprecated
from astropy import units as u
from astropy.nddata import (
    NDDataArray, StdDevUncertainty, VarianceUncertainty, InverseVariance
)
from specutils import Spectrum1D

from jdaviz.components.toolbar_nested import NestedJupyterToolbar
from jdaviz.configs.default.plugins.data_menu import DataMenu
from jdaviz.core.astrowidgets_api import AstrowidgetsImageViewerMixin
from jdaviz.core.custom_units_and_equivs import _eqv_sb_per_pixel_to_per_angle
from jdaviz.core.events import SnackbarMessage
from jdaviz.core.freezable_state import FreezableProfileViewerState
from jdaviz.core.marks import LineUncertainties, ScatterMask, OffscreenLinesMarks
from jdaviz.core.registries import viewer_registry
from jdaviz.core.template_mixin import WithCache
from jdaviz.core.user_api import ViewerUserApi
from jdaviz.core.unit_conversion_utils import check_if_unit_is_per_solid_angle
from jdaviz.utils import (ColorCycler, get_subset_type, _wcs_only_label,
                          layer_is_image_data, layer_is_not_dq)

uc = UnitConverter()

uncertainty_str_to_cls_mapping = {
    "std": StdDevUncertainty,
    "var": VarianceUncertainty,
    "ivar": InverseVariance
}


__all__ = ['JdavizViewerMixin', 'JdavizProfileView']

viewer_registry.add("g-profile-viewer", label="Profile 1D", cls=BqplotProfileView)
viewer_registry.add("g-image-viewer", label="Image 2D", cls=BqplotImageView)
viewer_registry.add("g-table-viewer", label="Table", cls=TableViewer)


class JdavizViewerMixin(WithCache):
    toolbar = None
    tools_nested = []
    _prev_limits = None
    _native_mark_classnames = ('Lines', 'LinesGL', 'FRBImage', 'Contour')

    def __init__(self, *args, **kwargs):
        # NOTE: anything here most likely won't be called by viewers because of inheritance order
        super().__init__(*args, **kwargs)

        # Allow each viewer to cycle through colors for each new addition to the viewer:
        self.color_cycler = ColorCycler()

        self._data_menu = DataMenu(viewer=self, app=self.jdaviz_app)

    @property
    def user_api(self):
        # default exposed user APIs.  Can override this method in any particular viewer.
        if not isinstance(self, TableViewer):
            # TODO: eventually remove data_labels_loaded
            # and data_labels_visible once deprecation period passes
            expose = ['data_labels_loaded', 'data_labels_visible', 'data_menu']
        else:
            expose = []
        if isinstance(self, BqplotImageView):
            if isinstance(self, AstrowidgetsImageViewerMixin):
                expose += ['save',
                           'center_on', 'offset_by', 'zoom_level', 'zoom',
                           'colormap_options', 'set_colormap',
                           'stretch_options', 'stretch',
                           'autocut_options', 'cuts',
                           'marker', 'add_markers', 'remove_markers', 'reset_markers',
                           'blink_once', 'reset_limits']
            else:
                # cubeviz image viewers don't inherit from AstrowidgetsImageViewerMixin yet,
                # but also shouldn't expose set_limits because of equal aspect ratio concerns
                expose += []
        elif isinstance(self, TableViewer):
            expose += []
        else:
            expose += ['set_limits', 'reset_limits', 'set_tick_format']
        return ViewerUserApi(self, expose=expose)

    @property
    def data_menu(self):
        return self._data_menu.user_api

    def _deprecated_data_menu(self):
        # temporary method to allow for opening new data-menu from old button.  This should
        # be removed anytime after the old button is removed (likely in 4.3)
        self.data_menu.open_menu()

    @property
    @deprecated(since="4.1", alternative="viewer.data_menu.data_labels_loaded")
    def data_labels_loaded(self):
        """
        List of data labels loaded in this viewer.

        Returns
        -------
        data_labels : list
            list of strings
        """
        return self.data_menu.data_labels_loaded

    @property
    @deprecated(since="4.1", alternative="viewer.data_menu.data_labels_visible")
    def data_labels_visible(self):
        """
        List of data labels visible in this viewer.

        Returns
        -------
        data_labels : list
            list of strings
        """
        return self.data_menu.data_labels_visible

    def reset_limits(self):
        """
        Reset viewer axes limits.
        """
        self.state.reset_limits()

    def set_limits(self, x_min=None, x_max=None, y_min=None, y_max=None):
        """
        Set viewer axes limits.

        Parameters
        ----------
        x_min : float or None, optional
            lower-limit of x-axis (in current axes units)
        x_max: float or None, optional
            upper-limit of x-axis (in current axes units)
        y_min : float or None, optional
            lower-limit of y-axis (in current axes units)
        y_max: float or None, optional
            upper-limit of y-axis (in current axes units)
        """
        for val in (x_min, x_max, y_min, y_max):
            if val is not None and not isinstance(val, (float, int, np.float32)):
                raise TypeError('All arguments must be None, int, or float, '
                                f'but got: {type(val)}')

        with delay_callback(self.state, 'x_min', 'x_max', 'y_min', 'y_max'):
            if x_min is not None:
                self.state.x_min = x_min
            if x_max is not None:
                self.state.x_max = x_max
            # NOTE: for some reason, setting ymax first avoids an issue
            # where back-to-back calls of get_limits and set_limits
            # give different results for y limits.
            if y_max is not None:
                self.state.y_max = y_max
            if y_min is not None:
                self.state.y_min = y_min

    def get_limits(self):
        """Return current viewer axes limits.

        Returns
        -------
        x_min, x_max, y_min, y_max : float
            Lower/upper X/Y limits, respectively.

        """
        return self.state.x_min, self.state.x_max, self.state.y_min, self.state.y_max

    def set_tick_format(self, fmt, axis):
        """
        Manually set the tick format of one of the axes.

        Parameters
        ----------
        fmt : str
            Format of tick marks.
            For example, ``'0.1e'`` to set scientific notation or ``'0.2f'`` to turn it off.
        axis : {x, y}
            The viewer axis.
        """
        if axis not in ('x', 'y'):
            raise ValueError("axis must be 'x' or 'y'")
        # Examples of values for fmt are '0.1e' or '0.2f'
        axis = {'x': 0, 'y': 1}[axis]
        self.figure.axes[axis].tick_format = fmt

    @property
    def native_marks(self):
        """
        Return all marks that are Lines/LinesGL objects (and not subclasses)
        """
        return [m for m in self.figure.marks
                if m.__class__.__name__ in self._native_mark_classnames]

    @property
    def custom_marks(self):
        """
        Return all marks that are not Lines/LinesGL objects (but can be subclasses)
        """
        return [m for m in self.figure.marks
                if m.__class__.__name__ not in self._native_mark_classnames]

    def _subscribe_to_layers_update(self):
        # subscribe to new layers
        self._expected_subset_layers = []
        self._layers_with_defaults_applied = []
        self.state.add_callback('layers', self._on_layers_update)

    def _get_layer(self, label):
        for layer in self.state.layers:
            if layer.layer.label == label:
                return layer

    def _apply_layer_defaults(self, layer_state):
        if hasattr(layer_state, 'as_steps'):
            if layer_state.layer.label != layer_state.layer.data.label:
                # then this is a subset, so default based on the parent data layer
                layer_state.as_steps = self._get_layer(layer_state.layer.data.label).as_steps
            else:
                # default to not plotting with as_steps (despite glue defaulting to True)
                layer_state.as_steps = False
            # whenever as_steps changes, we need to redraw the uncertainties (if enabled)
            layer_state.add_callback('as_steps', self._show_uncertainty_changed)

    def _expected_subset_layer_default(self, layer_state):
        if self.__class__.__name__ == 'RampvizImageView':
            # Do not override default for subsets as for some reason
            # this isn't getting called when they're first added, but rather when
            # the next state change is made (for example: manually changing the visibility)
            return

        # default visibility based on the visibility of the "parent" data layer
        if self.__class__.__name__ == 'RampvizProfileView':
            # Rampviz doesn't show subset profiles by default:
            layer_state.visible = False
        elif (self.__class__.__name__ == 'CubevizImageView' and
              get_subset_type(layer_state.layer) != 'spatial'):
            # set visibility of spectral subsets to false in Cubeviz image-viewers
            layer_state.visible = False
        else:
            layer_state.visible = self._get_layer(layer_state.layer.data.label).visible

    def _update_layer_icons(self):
        # update visible_layers (TODO: move this somewhere that can update on color change, etc)
        def _get_layer_color(layer):
            if isinstance(layer, BqplotScatterLayerState):
                # then this could be a scatter layer in an image viewer,
                # so we'll ignore the color_mode
                return layer.color
            if getattr(self.state, 'color_mode', None) == 'Colormaps':
                for subset in self.jdaviz_app.data_collection.subset_groups:
                    if subset.label == layer.layer.label:
                        # then we still want to show the color for a subset
                        return layer.color
                # then this is a data-layer in colormap mode, so we'll ignore the color
                return ''
            return layer.color

        def _get_layer_linewidth(layer):
            linewidth = getattr(layer, 'linewidth', 0)
            return min(linewidth, 6)

        def _get_layer_info(layer):
            if 'Trace' in layer.layer.data.meta:
                return "mdi-chart-line-stacked", None

            for subset in self.jdaviz_app.data_collection.subset_groups:
                if subset.label == layer.layer.label:
                    subset_type = get_subset_type(subset)
                    if subset_type == 'spatial':
                        return "mdi-chart-scatter-plot", subset_type
                    else:
                        return "mdi-chart-bell-curve", subset_type
            return "", None

        visible_layers = {}
        for layer in self.state.layers[::-1]:
            layer_is_wcs_only = (
                    hasattr(layer.layer, 'meta') and
                    layer.layer.meta.get(_wcs_only_label, False)
            )
            if layer.visible and not layer_is_wcs_only:
                prefix_icon, subset_type = _get_layer_info(layer)
                if (
                    subset_type == 'spatial' and
                    self.__class__.__name__ in ('CubevizProfileView', 'RampvizProfileView')
                ):
                    # do not show spatial subsets in profile viewer
                    continue
                visible_layers[layer.layer.label] = {'color': _get_layer_color(layer),
                                                     'linewidth': _get_layer_linewidth(layer),
                                                     'prefix_icon': prefix_icon}

        self._data_menu.visible_layers = visible_layers

    def _on_layers_update(self, layers=None):
        if self.__class__.__name__ == 'MosvizTableViewer':
            # MosvizTableViewer uses this as a mixin, but we do not need any of this layer
            # logic there
            return

        viewer_item = self.jdaviz_app._viewer_item_by_id(self.reference_id)
        if viewer_item is None:
            return
        selected_data_items = viewer_item.get('selected_data_items', {})

        # update selected_data_items
        for data_id, visibility in selected_data_items.items():
            label = next((x['name'] for x in self.jdaviz_app.state.data_items
                          if x['id'] == data_id), None)

            visibilities = []
            for layer in self.state.layers:
                if layer.layer.data.label == label:
                    visibilities.append(layer.visible)
            if np.all(visibilities):
                selected_data_items[data_id] = 'visible'
            elif np.any(visibilities):
                selected_data_items[data_id] = 'mixed'
            else:
                selected_data_items[data_id] = 'hidden'

        self._update_layer_icons()

        # we'll make a deepcopy so that we can remove entries from the self._expected_subset_layers
        # to avoid recursion, but also handle multiple layers for the same subset
        expected_subset_layers = self._expected_subset_layers[:]
        for layer in self.state.layers:
            layer_info = {'data_label': layer.layer.data.label,
                          'layer_label': layer.layer.label}
            if layer_info not in self._layers_with_defaults_applied:
                self._layers_with_defaults_applied.append(layer_info)
                self._apply_layer_defaults(layer)

            if layer.layer.label in expected_subset_layers:
                if layer.layer.label in self._expected_subset_layers:
                    self._expected_subset_layers.remove(layer.layer.label)
                self._expected_subset_layer_default(layer)

    def _on_subset_create(self, msg):
        from jdaviz.configs.mosviz.plugins.viewers import MosvizTableViewer
        if isinstance(self, MosvizTableViewer):
            # MosvizTableViewer uses this as a mixin, but we do not need any of this layer
            # logic there
            return

        # NOTE: the subscription to this method is handled in ConfigHelper
        # we don't have access to the actual subset yet to tell if its spectral or spatial, so
        # we'll store the name of this new subset and change the default linewidth when the
        # layers are added
        if msg.subset.label not in self._expected_subset_layers and msg.subset.label:
            self._expected_subset_layers.append(msg.subset.label)

    def _on_subset_delete(self, msg):
        """
        This is needed to remove the "ghost" subset left over when the subset tool is active,
        and the active subset is deleted. https://github.com/spacetelescope/jdaviz/issues/2499
        is open to revert/update this if it ends up being addressed upstream in
        https://github.com/glue-viz/glue-jupyter/issues/401.
        """
        from jdaviz.configs.mosviz.plugins.viewers import MosvizTableViewer
        if isinstance(self, MosvizTableViewer):
            # MosvizTableViewer uses this as a mixin, but we do not need any of this layer
            # logic there
            return

        subset_tools = ['bqplot:truecircle', 'bqplot:rectangle', 'bqplot:ellipse',
                        'bqplot:circannulus', 'bqplot:xrange']

        if not len(self.session.edit_subset_mode.edit_subset):
            if self.toolbar.active_tool_id in subset_tools:
                if (hasattr(self.toolbar, "default_tool_priority") and
                        len(self.toolbar.default_tool_priority)):
                    self.toolbar.active_tool_id = self.toolbar.default_tool_priority[0]
                else:
                    self.toolbar.active_tool = None

    @property
    def active_image_layer(self):
        """Active image layer in the viewer, if available."""
        # Find visible layers
        visible_layers = [layer for layer in self.state.layers
                          if (layer.visible and
                              layer_is_image_data(layer.layer) and
                              layer_is_not_dq(layer.layer) and
                              (layer.bitmap_visible or layer.contour_visible))]
        if len(visible_layers) == 0:
            return None

        return visible_layers[-1]

    def initialize_toolbar(self, default_tool_priority=[]):
        # NOTE: this overrides glue_jupyter.IPyWidgetView
        self.toolbar = NestedJupyterToolbar(self, self.tools_nested, default_tool_priority)

    @property
    def tools(self):
        # NOTE: this overrides the default list of tools for the BasicJupyterToolbar by
        # returning a flattened version of self.tools_nested
        return list(self.toolbar.tools.keys())

    @property
    def jdaviz_app(self):
        """The Jdaviz application tied to the viewer."""
        return self.session.jdaviz_app

    @property
    def jdaviz_helper(self):
        """The Jdaviz configuration helper tied to the viewer."""
        return self.jdaviz_app._jdaviz_helper

    @property
    def hub(self):
        return self.session.hub

    @property
    def reference_id(self):
        return self._reference_id

    @property
    def reference(self):
        return self.jdaviz_app._viewer_item_by_id(self.reference_id).get('reference')

    @property
    def _ref_or_id(self):
        reference = self.reference
        if reference is not None:
            return reference
        return self.reference_id

    def set_plot_axes(self):
        # individual viewers can override to set custom axes labels/ticks/styling
        return


@viewer_registry("jdaviz-profile-viewer", label="Profile 1D")
class JdavizProfileView(JdavizViewerMixin, BqplotProfileView):
    # categories: zoom resets, zoom, pan, subset, select tools, shortcuts
    tools_nested = [
                    ['jdaviz:homezoom', 'jdaviz:prevzoom'],
                    ['jdaviz:boxzoom', 'jdaviz:xrangezoom', 'jdaviz:yrangezoom'],
                    ['jdaviz:panzoom', 'jdaviz:panzoom_x', 'jdaviz:panzoom_y'],
                    ['bqplot:xrange'],
                    ['jdaviz:sidebar_plot', 'jdaviz:sidebar_export']
                ]

    default_class = NDDataArray
    _state_cls = FreezableProfileViewerState
    _default_profile_subset_type = None

    def __init__(self, *args, **kwargs):
        default_tool_priority = kwargs.pop('default_tool_priority', [])
        super().__init__(*args, **kwargs)

        self._subscribe_to_layers_update()
        self.initialize_toolbar(default_tool_priority=default_tool_priority)
        self._offscreen_lines_marks = OffscreenLinesMarks(self)
        self.figure.marks = self.figure.marks + self._offscreen_lines_marks.marks

        self.state.add_callback('show_uncertainty', self._show_uncertainty_changed)

        self.display_mask = False

        # Change collapse function to sum
        default_collapse_function = kwargs.pop('default_collapse_function', 'sum')

        self.state.function = default_collapse_function

    def _expected_subset_layer_default(self, layer_state):
        super()._expected_subset_layer_default(layer_state)

        layer_state.linewidth = 3

    def data(self, cls=None):
        # Grab the user's chosen statistic for collapsing data
        statistic = getattr(self.state, 'function', None)
        data = []

        for layer_state in self.state.layers:
            if hasattr(layer_state, 'layer'):
                lyr = layer_state.layer

                # For raw data, just include the data itself
                if isinstance(lyr, BaseData):
                    _class = cls or self.default_class

                    if _class is not None:
                        cache_key = (lyr.label, statistic)
                        if cache_key in self.jdaviz_app._get_object_cache:
                            layer_data = self.jdaviz_app._get_object_cache[cache_key]
                        else:
                            # If spectrum, collapse via the defined statistic
                            if _class == Spectrum1D:
                                layer_data = lyr.get_object(cls=_class, statistic=statistic)
                            else:
                                layer_data = lyr.get_object(cls=_class)
                            self.jdaviz_app._get_object_cache[cache_key] = layer_data

                        data.append(layer_data)

                # For subsets, make sure to apply the subset mask to the layer data first
                elif isinstance(lyr, Subset):
                    layer_data = lyr

                    if _class is not None:
                        handler, _ = data_translator.get_handler_for(_class)
                        try:
                            layer_data = handler.to_object(layer_data, statistic=statistic)
                        except IncompatibleAttribute:
                            continue
                    data.append(layer_data)

        return data

    def get_scales(self):
        fig = self.figure
        # Deselect any pan/zoom or subsetting tools so they don't interfere
        # with the scale retrieval
        if self.toolbar.active_tool is not None:
            self.toolbar.active_tool = None
        return {'x': fig.interaction.x_scale, 'y': fig.interaction.y_scale}

    def _show_uncertainty_changed(self, msg=None):
        # this is subscribed in init to watch for changes to the state
        # object since uncertainty handling is in jdaviz instead of glue/glue-jupyter
        if self.state.show_uncertainty:
            self._plot_uncertainties()
        else:
            self._clean_error()

    def show_mask(self):
        self.display_mask = True
        self._plot_mask()

    def clean(self):
        # Remove extra traces, in case they exist.
        self.display_mask = False
        self._clean_mask()

        # this will automatically call _clean_error via _show_uncertainty_changed
        self.state.show_uncertainty = False

    def _clean_mask(self):
        fig = self.figure
        fig.marks = [x for x in fig.marks if not isinstance(x, ScatterMask)]

    def _clean_error(self):
        fig = self.figure
        fig.marks = [x for x in fig.marks if not isinstance(x, LineUncertainties)]

    def add_data(self, data, color=None, alpha=None, **layer_state):
        """
        Overrides the base class to add markers for plotting
        uncertainties and data quality flags.

        Parameters
        ----------
        spectrum : :class:`glue.core.data.Data`
            Data object with the spectrum.
        color : obj
            Color value for plotting.
        alpha : float
            Alpha value for plotting.

        Returns
        -------
        result : bool
            `True` if successful, `False` otherwise.
        """
        # If this is the first loaded data, set things up for unit conversion.
        if len(self.layers) == 0:
            reset_plot_axes = True
        else:
            # Check if the new data flux unit is actually compatible since flux not linked.
            try:
                if self.state.y_display_unit not in ['None', None, 'DN']:
                    uc.to_unit(data, data.find_component_id("flux"), [1, 1],
                               u.Unit(self.state.y_display_unit))  # Error if incompatible
            except Exception as err:
                # Raising exception here introduces a dirty state that messes up next load_data
                # but not raising exception also causes weird behavior unless we remove the data
                # completely.
                self.session.hub.broadcast(SnackbarMessage(
                    f"Failed to load {data.label}, so removed it: {repr(err)}",
                    sender=self, color='error'))
                self.jdaviz_app.data_collection.remove(data)
                return False
            reset_plot_axes = False

        # The base class handles the plotting of the main
        # trace representing the profile itself.
        result = super().add_data(data, color, alpha, **layer_state)

        if reset_plot_axes:
            x_units = data.get_component(self.state.x_att.label).units

            y_axis_component = (
                'flux' if 'flux' in [comp.label for comp in self.state.layers[0].layer.components]
                else 'data'
            )
            y_units = data.get_component(y_axis_component).units
            with delay_callback(self.state, "x_display_unit", "y_display_unit"):
                self.state.x_display_unit = x_units if len(x_units) else None
                self.state.y_display_unit = y_units if len(y_units) else None
            self.set_plot_axes()

        self._plot_uncertainties()

        self._plot_mask()

        # Set default linewidth on any created spectral subset layers
        # NOTE: this logic will need updating if we add support for multiple cubes as this assumes
        # that new data entries (from model fitting or gaussian smooth, etc) will only be spectra
        # and all subsets affected will be spectral
        for layer in self.state.layers:
            if (isinstance(layer.layer, GroupedSubset)
                    and get_subset_type(layer.layer) == self._default_profile_subset_type
                    and layer.layer.data.label == data.label):
                layer.linewidth = 3

        return result

    def _plot_mask(self):
        if not self.display_mask:
            return

        # Remove existing mask marks
        self._clean_mask()

        # Loop through all active data in the viewer
        for index, layer_state in enumerate(self.state.layers):
            lyr = layer_state.layer
            comps = [str(component) for component in lyr.components]

            # Skip subsets
            if hasattr(lyr, "subset_state"):
                continue

            # Ignore data that does not have a mask component
            if "mask" in comps:
                mask = np.array(lyr['mask'].data)

                data_obj = lyr.data.get_object(cls=self.default_class)

                if self.default_class == Spectrum1D:
                    data_x = data_obj.spectral_axis.value
                    data_y = data_obj.flux.value
                else:
                    data_x = np.arange(data_obj.shape[-1])
                    data_y = data_obj.data.value

                # For plotting markers only for the masked data
                # points, erase un-masked data from trace.
                y = np.where(np.asarray(mask) == 0, np.nan, data_y)

                # A subclass of the bqplot Scatter object, ScatterMask places
                # 'X' marks where there is masked data in the viewer.
                color = layer_state.color
                alpha_shade = layer_state.alpha / 3
                mask_line_mark = ScatterMask(scales=self.scales,
                                             marker='cross',
                                             x=data_x,
                                             y=y,
                                             stroke_width=0.5,
                                             colors=[color],
                                             default_size=25,
                                             default_opacities=[alpha_shade]
                                             )
                # Add mask marks to viewer
                self.figure.marks = list(self.figure.marks) + [mask_line_mark]

    def _plot_uncertainties(self):
        if not self.state.show_uncertainty:
            return

        # Remove existing error bars
        self._clean_error()

        # Loop through all active data in the viewer
        for index, layer_state in enumerate(self.state.layers):
            lyr = layer_state.layer

            # Skip subsets
            if hasattr(lyr, "subset_state"):
                continue

            comps = [str(component) for component in lyr.components]

            # Ignore data that does not have an uncertainty component
            if "uncertainty" in comps:  # noqa
                error = np.array(lyr['uncertainty'].data)

                # ensure that the uncertainties are represented as stddev:
                uncertainty_type_str = lyr.meta.get('uncertainty_type', 'stddev')
                uncert_cls = uncertainty_str_to_cls_mapping[uncertainty_type_str]
                error = uncert_cls(error).represent_as(StdDevUncertainty).array

                # Then we assume that last axis is always wavelength.
                # This may need adjustment after the following
                # specutils PR is merged: https://github.com/astropy/specutils/pull/1033
                spectral_axis = -1
                data_obj = lyr.data.get_object(cls=self.default_class, statistic=None)

                if isinstance(lyr.data.coords, SpectralCoordinates):
                    spectral_wcs = lyr.data.coords
                    data_x = spectral_wcs.pixel_to_world_values(
                        np.arange(lyr.data.shape[spectral_axis])
                    )
                    if isinstance(data_x, tuple):
                        data_x = data_x[0]
                else:
                    if hasattr(lyr.data.coords, 'spectral_wcs'):
                        spectral_wcs = lyr.data.coords.spectral_wcs
                    elif hasattr(lyr.data.coords, 'spectral'):
                        spectral_wcs = lyr.data.coords.spectral
                    data_x = spectral_wcs.pixel_to_world(
                        np.arange(lyr.data.shape[spectral_axis])
                    )

                data_y = data_obj.data

                # The shaded band around the spectrum trace is bounded by
                # two lines, above and below the spectrum trace itself.
                data_x_list = np.ndarray.tolist(data_x)
                x = [data_x_list, data_x_list]
                y = [np.ndarray.tolist(data_y - error),
                     np.ndarray.tolist(data_y + error)]

                if layer_state.as_steps:
                    for i in (0, 1):
                        a = np.insert(x[i], 0, 2*x[i][0] - x[i][1])
                        b = np.append(x[i], 2*x[i][-1] - x[i][-2])
                        edges = (a + b) / 2
                        x[i] = np.concatenate((edges[:1], np.repeat(edges[1:-1], 2), edges[-1:]))
                        y[i] = np.repeat(y[i], 2)
                x, y = np.asarray(x), np.asarray(y)

                # A subclass of the bqplot Lines object, LineUncertainties keeps
                # track of uncertainties plotted in the viewer. LineUncertainties
                # appear with two lines and shaded area in between.
                color = layer_state.color
                alpha_shade = layer_state.alpha / 3
                error_line_mark = LineUncertainties(viewer=self,
                                                    x=[x],
                                                    y=[y],
                                                    scales=self.scales,
                                                    stroke_width=1,
                                                    colors=[color, color],
                                                    fill_colors=[color, color],
                                                    opacities=[0.0, 0.0],
                                                    fill_opacities=[alpha_shade,
                                                                    alpha_shade],
                                                    fill='between',
                                                    close_path=False
                                                    )

                # Add error lines to viewer
                self.figure.marks = list(self.figure.marks) + [error_line_mark]

    def set_plot_axes(self):
        # Set x and y axes labels for the spectrum viewer
        y_display_unit = self.state.y_display_unit
        y_unit = (
            u.Unit(y_display_unit) if y_display_unit and y_display_unit != 'None'
            else u.dimensionless_unscaled
        )

        # Get local units.
        locally_defined_flux_units = [
            u.Jy, u.mJy, u.uJy, u.MJy,
            u.W / (u.m**2 * u.Hz),
            u.eV / (u.s * u.m**2 * u.Hz),
            u.erg / (u.s * u.cm**2),
            u.erg / (u.s * u.cm**2 * u.Angstrom),
            u.erg / (u.s * u.cm**2 * u.Hz),
            u.ph / (u.s * u.cm**2 * u.Angstrom),
            u.ph / (u.s * u.cm**2 * u.Hz),
            u.bol, u.AB, u.ST
        ]

        # get square angle from 'sb' display unit
        sb_unit = self.jdaviz_app._get_display_unit(axis='sb')
        if sb_unit is not None:
            solid_angle_unit = check_if_unit_is_per_solid_angle(sb_unit, return_unit=True)
        else:
            solid_angle_unit = None

        # if solid angle is present in denominator, check physical type of numerator
        # if numerator is a flux type the display unit is a 'surface brightness', otherwise
        # default to the catchall 'flux density' label
        flux_unit_type = None

        for un in locally_defined_flux_units:
            locally_defined_sb_unit = un / solid_angle_unit if solid_angle_unit is not None else None  # noqa

            # create an equivalency for each flux unit for flux <> flux/pix2.
            # for similar reasons to the 'untranslatable units' issue, custom
            # equivs. can't be combined, so a workaround is creating an eqiv
            # for each flux that may need an additional equiv.
            angle_to_pixel_equiv = _eqv_sb_per_pixel_to_per_angle(un)

            if (locally_defined_sb_unit is not None
                    and y_unit.is_equivalent(locally_defined_sb_unit, angle_to_pixel_equiv)):
                flux_unit_type = "Surface Brightness"
            elif y_unit.is_equivalent(un):
                flux_unit_type = 'Flux'
            elif y_unit.is_equivalent(u.electron / u.s) or y_unit.physical_type == 'dimensionless':  # noqa
                # electron / s or 'dimensionless_unscaled' should be labeled counts
                flux_unit_type = "Counts"
            elif y_unit.is_equivalent(u.W):
                flux_unit_type = "Luminosity"
            if flux_unit_type is not None:
                # if we determined a label, stop checking
                break
        else:
            # default to Flux Density for flux density or uncaught types
            flux_unit_type = "Flux density"

        # Set x axes labels for the spectrum viewer
        x_disp_unit = self.state.x_display_unit
        x_unit = u.Unit(x_disp_unit) if x_disp_unit else u.dimensionless_unscaled

        if x_unit.is_equivalent(u.m):
            spectral_axis_unit_type = "Wavelength"
        elif x_unit.is_equivalent(u.Hz):
            spectral_axis_unit_type = "Frequency"
        elif x_unit.is_equivalent(u.pixel):
            spectral_axis_unit_type = "Pixel"
        elif x_unit.is_equivalent(u.dimensionless_unscaled):
            # case for rampviz
            spectral_axis_unit_type = "Group"
        else:
            spectral_axis_unit_type = str(x_unit.physical_type).title()

        with self.figure.hold_sync():
            self.figure.axes[0].label = f"{spectral_axis_unit_type}" + (
                f" [{self.state.x_display_unit}]"
                if self.state.x_display_unit not in ["None", None] else ""
            )
            self.figure.axes[1].label = f"{flux_unit_type}" + (
                f"[{self.state.y_display_unit}]"
                if self.state.y_display_unit not in ["None", None] else ""
            )

            # Make it so axis labels are not covering tick numbers.
            self.figure.fig_margin["left"] = 95
            self.figure.fig_margin["bottom"] = 60
            self.figure.send_state('fig_margin')  # Force update
            self.figure.axes[0].label_offset = "40"
            self.figure.axes[1].label_offset = "-70"
            # NOTE: with tick_style changed below, the default responsive ticks in bqplot result
            # in overlapping tick labels.  For now we'll hardcode at 8, but this could be removed
            # (default to None) if/when bqplot auto ticks react to styling options.
            self.figure.axes[1].num_ticks = 8

            # Set Y-axis to scientific notation
            self.figure.axes[1].tick_format = '0.1e'

            for i in (0, 1):
                self.figure.axes[i].tick_style = {'font-size': 15, 'font-weight': 600}



import numpy as np
from numpy.testing import assert_allclose
import pytest

import astropy.units as u
from astropy.table import QTable
from specutils import Spectrum1D

from jdaviz.core.marks import SpectralLine
from jdaviz.core.linelists import get_available_linelists


def test_line_lists(specviz_helper):
    spec = Spectrum1D(flux=np.random.rand(100)*u.Jy,
                      spectral_axis=np.arange(6000, 7000, 10)*u.AA)
    specviz_helper.load_data(spec)

    lt = QTable()
    lt['linename'] = ['O III', 'Halpha']
    lt['rest'] = [0, 6563]*u.AA
    with pytest.raises(ValueError, match='all rest values must be positive'):
        specviz_helper.load_line_list(lt)

    lt = QTable()
    lt['linename'] = ['O III', 'Halpha']
    lt['rest'] = [5007, 6563]*u.AA
    specviz_helper.load_line_list(lt)

    assert len(specviz_helper.spectral_lines) == 2
    assert specviz_helper.spectral_lines.loc["linename", "Halpha"]["listname"] == "Custom"
    assert np.all(specviz_helper.spectral_lines["show"])
    assert specviz_helper.plugins['Line Lists']._obj.rs_enabled is True

    specviz_helper.erase_spectral_lines()
    assert np.all(specviz_helper.spectral_lines["show"] == False)  # noqa
    assert specviz_helper.plugins['Line Lists']._obj.rs_enabled is False

    specviz_helper.plot_spectral_line("Halpha")
    specviz_helper.plot_spectral_line("O III 5007.0")

    assert np.all(specviz_helper.spectral_lines["show"])

    assert (
        specviz_helper.plugins["Line Lists"]._obj.list_contents["Custom"]["medium"]
        == "Unknown (Custom)"
    )


def test_redshift(specviz_helper, spectrum1d):
    # Also test that plugin is disabled before data is loaded.
    ll_plugin = specviz_helper.plugins['Line Lists']._obj
    assert ll_plugin.disabled_msg

    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    assert not ll_plugin.disabled_msg

    lt = QTable()
    lt['linename'] = ['O III', 'Halpha']
    lt['rest'] = [5007, 6563]*u.AA
    lt['redshift'] = u.Quantity(0.046)
    lt['listname'] = 'Test List'
    with pytest.warns(UserWarning, match='per line/list redshifts not supported, use viz.set_redshift'):  # noqa
        specviz_helper.load_line_list(lt)

    # open the plugin so that all updates run
    ll_plugin.plugin_opened = True
    line = ll_plugin.list_contents['Test List']['lines'][0]
    assert_allclose(line['obs'], line['rest'])
    # test API access
    specviz_helper.set_redshift(0.01)
    specviz_helper.set_redshift_slider_bounds(range=0.5, step=0.01)
    specviz_helper.set_redshift_slider_bounds(range='auto', step='auto')

    # test plugin
    ll_plugin.rs_redshift = 0.1
    assert ll_plugin.rs_rv == 28487.06614479641

    ll_plugin.rs_rv = 30000
    assert ll_plugin.rs_redshift == 0.10561890816244568

    # https://github.com/spacetelescope/jdaviz/issues/1692
    # adding new data entry from a plugin should not reset redshift
    specviz_helper.plugins['Gaussian Smooth'].smooth()
    assert ll_plugin.rs_redshift == 0.10561890816244568

    # test that setting observed wavelength works
    ll_plugin.vue_change_line_obs({'list_name': 'Test List',
                                   'line_ind': 0,
                                   'obs_new': 5508})
    assert_allclose(line['obs'], 5508)
    assert ll_plugin.rs_redshift == 0.10005991611743559

    # https://github.com/spacetelescope/jdaviz/issues/1168
    ll_plugin.vue_set_identify(('Test List', line, 0))
    ll_plugin.vue_remove_list('Test List')
    assert ll_plugin._viewer.spectral_lines is None
    assert ll_plugin.identify_label == ''


def test_load_available_preset_lists(specviz_helper, spectrum1d):
    """ Loads all available line lists and checks the medium requirement """
    label = "Test 1D Spectrum"
    specviz_helper.load_data(spectrum1d, data_label=label)

    # Check to make sure we got our line lists
    available_linelists = get_available_linelists()
    assert len(available_linelists) > 0

    for linelist in available_linelists:
        specviz_helper.plugins['Line Lists']._obj.vue_list_selected(linelist)
        specviz_helper.plugins['Line Lists']._obj.vue_load_list(linelist)

    # Check that we loaded all the lists (+1 because of the Custom list)
    assert (
        len(specviz_helper.plugins['Line Lists']._obj.list_contents.keys()) ==
        len(available_linelists) + 1
    )

    # Line list must have "medium" info to be available
    for list in specviz_helper.plugins['Line Lists']._obj.list_contents.values():  # noqa
        assert 'medium' in list


def test_line_identify(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)

    lt = QTable()
    lt['linename'] = ['O III', 'Halpha']
    lt['rest'] = [5007, 6563]*u.AA
    lt['listname'] = 'Test List'
    specviz_helper.load_line_list(lt)

    ll_plugin = specviz_helper.app.get_tray_item_from_name('g-line-list')
    line = ll_plugin.list_contents['Test List']['lines'][0]
    assert line.get('identify', False) is False
    ll_plugin.vue_set_identify(('Test List', line, 0))
    assert line.get('identify', False) is True

    ll_plugin.vue_change_visible(('Test List', line, 0))
    assert line.get('show') is False
    assert line.get('identify', False) is False


def test_global_redshift_applied(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)

    # Create a table with redshift included
    lt = QTable({'linename': ['O III', 'Halpha'],
                 'rest': [5007, 6563] * u.AA,
                 'redshift': u.Quantity([0, 0])})

    with pytest.warns(UserWarning, match='per line/list redshifts not supported, use viz.set_redshift'):  # noqa
        specviz_helper.load_line_list(lt)

    # Load a line, and apply redshift globally
    specviz_helper.plot_spectral_line("Halpha")
    specviz_helper.set_redshift(0.01)
    # Load second line, redshift should also be applied to it
    specviz_helper.plot_spectral_line("O III")

    viewer_lines = [mark for mark in specviz_helper.app.get_viewer(
        specviz_helper._default_spectrum_viewer_reference_name).figure.marks
        if isinstance(mark, SpectralLine)]

    assert np.allclose([line.redshift for line in viewer_lines], 0.01)


def test_global_redshift_applied_to_all(specviz_helper, spectrum1d):
    specviz_helper.load_data(spectrum1d)

    # Create a table with redshift included
    lt = QTable({'linename': ['O III', 'Halpha', 'O I'],
                 'rest': [5007, 6563, 6300] * u.AA,
                 'redshift': u.Quantity([0, 0, 0])})

    with pytest.warns(UserWarning, match='per line/list redshifts not supported, use viz.set_redshift'):  # noqa
        specviz_helper.load_line_list(lt)

    # Load a line, so we can apply redshift
    specviz_helper.plot_spectral_line("Halpha")
    global_redshift = 0.01
    specviz_helper.set_redshift(global_redshift)
    # Load remaining lines
    specviz_helper.plot_spectral_lines(global_redshift)

    viewer_lines = [mark for mark in specviz_helper.app.get_viewer(
        specviz_helper._default_spectrum_viewer_reference_name).figure.marks
        if isinstance(mark, SpectralLine)]

    assert np.allclose([line.redshift for line in viewer_lines], 0.01)



import numpy as np
from astropy import units as u
from traitlets import Bool, observe

from jdaviz.core.events import (ViewerAddedMessage, ChangeRefDataMessage,
                                AddDataMessage, RemoveDataMessage,
                                MarkersPluginUpdate)
from jdaviz.core.marks import MarkersMark
from jdaviz.core.registries import tray_registry
from jdaviz.core.template_mixin import PluginTemplateMixin, ViewerSelectMixin, TableMixin
from jdaviz.core.user_api import PluginUserApi

__all__ = ['Markers']


@tray_registry('g-markers', label="Markers")
class Markers(PluginTemplateMixin, ViewerSelectMixin, TableMixin):
    """
    See the :ref:`Markers Plugin Documentation <markers-plugin>` for more details.

    Only the following attributes and methods are available through the
    :ref:`public plugin API <plugin-apis>`:

    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.show`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.open_in_tray`
    * :meth:`~jdaviz.core.template_mixin.PluginTemplateMixin.close_in_tray`
    * :meth:`~jdaviz.core.template_mixin.TableMixin.clear_table`
    * :meth:`~jdaviz.core.template_mixin.TableMixin.export_table`
    """
    template_file = __file__, "markers.vue"
    uses_active_status = Bool(True).tag(sync=True)

    _default_table_values = {'spectral_axis': np.nan,
                             'spectral_axis:unit': '',
                             'slice': np.nan,
                             'pixel_x': np.nan,
                             'pixel_y': np.nan,
                             'pixel:unreliable': None,
                             'world_ra': np.nan,
                             'world_dec': np.nan,
                             'world:unreliable': None,
                             'value': np.nan,
                             'value:unit': '',
                             'value:unreliable': None,
                             'index': np.nan}

    @property
    def user_api(self):
        return PluginUserApi(self, expose=('clear_table', 'export_table',))

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.config == 'cubeviz':
            headers = ['spectral_axis', 'spectral_axis:unit',
                       'slice', 'pixel_x', 'pixel_y',
                       'world_ra', 'world_dec', 'value', 'value:unit', 'viewer']

        elif self.config == 'imviz':
            headers = ['pixel_x', 'pixel_y', 'pixel:unreliable',
                       'world_ra', 'world_dec', 'world:unreliable',
                       'value', 'value:unit', 'value:unreliable',
                       'viewer']

        elif self.config == 'specviz':
            headers = ['spectral_axis', 'spectral_axis:unit',
                       'index', 'value', 'value:unit']

        elif self.config == 'specviz2d':
            # TODO: add "index" if/when specviz2d supports plotting spectral_axis
            headers = ['spectral_axis', 'spectral_axis:unit',
                       'pixel_x', 'pixel_y', 'value', 'value:unit', 'viewer']

        elif self.config == 'mosviz':
            headers = ['spectral_axis', 'spectral_axis:unit',
                       'pixel_x', 'pixel_y', 'world_ra', 'world_dec', 'index',
                       'value', 'value:unit', 'viewer']
        else:
            # allow downstream configs to override headers
            headers = kwargs.get('headers', [])

        headers += ['data_label']

        self.table.headers_avail = headers
        self.table.headers_visible = headers
        self.table._default_values_by_colname = self._default_table_values

        def clear_table_callback():
            for mark in self.marks.values():
                mark.clear()

            self.hub.broadcast(MarkersPluginUpdate(table_length=0, sender=self))

        self.table._clear_callback = clear_table_callback

        # subscribe to mouse events on any new viewers
        self.hub.subscribe(self, ViewerAddedMessage, handler=self._on_viewer_added)

        # account for image rotation due to a change in reference data
        self.hub.subscribe(self, ChangeRefDataMessage,
                           handler=lambda msg: self._recompute_mark_positions(msg.viewer))

        # enable/disable mark based on whether parent data entry is in viewer
        self.hub.subscribe(self, AddDataMessage,
                           handler=lambda msg: self._recompute_mark_positions(msg.viewer))

        self.hub.subscribe(self, RemoveDataMessage,
                           handler=lambda msg: self._recompute_mark_positions(msg.viewer))

        self.docs_description = "Press 'm' with the cursor over a viewer to log\
                                the mouseover information. To change the\
                                selected layer, click the layer cycler in the\
                                mouseover information section of the app-level\
                                toolbar."

        # description displayed under plugin title in tray
        self._plugin_description = 'Create markers on viewers.'

    def _create_viewer_callbacks(self, viewer):
        if not self.is_active:
            return

        callback = self._viewer_callback(viewer, self._on_viewer_key_event)
        viewer.add_event_callback(callback, events=['keydown'])

    def _on_viewer_added(self, msg):
        self._create_viewer_callbacks(self.app.get_viewer_by_id(msg.viewer_id))

    def _recompute_mark_positions(self, viewer):
        if self.table is None or self.table._qtable is None:
            return
        if 'world_ra' not in self.table.headers_avail:
            return

        viewer_id = viewer.reference if viewer.reference is not None else viewer.reference_id
        viewer_loaded_data = [lyr.layer.label for lyr in viewer.layers]
        data_labels = self.table._qtable['data_label']
        viewer_labels = self.table._qtable['viewer']
        # note: could eventually have a user-provided switch to show markers in other viewers
        # by just skipping this first viewer_label == viewer_id check
        in_viewer = [viewer_label == viewer_id and data_label in viewer_loaded_data
                     for viewer_label, data_label in zip(viewer_labels, data_labels)]

        viewer_mark = self._get_mark(viewer)
        if not np.any(in_viewer):
            viewer_mark.x, viewer_mark.y = [], []
            return

        orig_world_x = np.asarray(self.table._qtable['world_ra'][in_viewer])
        orig_world_y = np.asarray(self.table._qtable['world_dec'][in_viewer])

        if self.app._align_by.lower() == 'wcs':
            # convert from the sky coordinates in the table to pixels via the WCS of the current
            # reference data
            new_wcs = viewer.state.reference_data.coords
            try:
                new_x, new_y = new_wcs.world_to_pixel_values(orig_world_x*u.deg,
                                                             orig_world_y*u.deg)
            except Exception:
                # fail gracefully
                new_x, new_y = [], []
        elif self.app._align_by == 'pixels':
            # we need to convert based on the WCS of the individual data layers on which each mark
            # was first created
            new_x, new_y = np.zeros_like(orig_world_x), np.zeros_like(orig_world_y)
            for data_label in np.unique(data_labels[in_viewer]):
                these = data_labels[in_viewer] == data_label
                if not np.any(these):
                    continue
                wcs = self.app.data_collection[data_label].coords
                try:
                    new_x[these], new_y[these] = wcs.world_to_pixel_values(orig_world_x[these]*u.deg,  # noqa
                                                                           orig_world_y[these]*u.deg)  # noqa
                except Exception:
                    # fail gracefully
                    new_x, new_y = [], []
                    break
        else:
            raise NotImplementedError(f"align_by {self.app._align_by} not implemented")

        # check for entries that do not correspond to a layer or only have pixel coordinates
        pixel_only_inds = data_labels == ''
        if np.any(pixel_only_inds):
            # TODO: should we rescale these since pixel coordinates when linked by WCS are always
            # on the range 0-1 because of the orientation layer?  Or hide the pixel option in the
            # cycler when WCS-linked?
            pixel_x = np.asarray(self.table._qtable['pixel_x'])
            pixel_y = np.asarray(self.table._qtable['pixel_y'])
            new_x = np.append(new_x, pixel_x[pixel_only_inds])
            new_y = np.append(new_y, pixel_y[pixel_only_inds])

        viewer_mark.x, viewer_mark.y = new_x, new_y

    def _get_mark(self, viewer):
        matches = [mark for mark in viewer.figure.marks if isinstance(mark, MarkersMark)]
        if len(matches):
            return matches[0]
        mark = MarkersMark(viewer)
        viewer.figure.marks = viewer.figure.marks + [mark]
        return mark

    @property
    def marks(self):
        return {viewer_id: self._get_mark(viewer)
                for viewer_id, viewer in self.app._viewer_store.items()
                if hasattr(viewer, 'figure')}

    @property
    def coords_info(self):
        return self.app.session.application._tools['g-coords-info']

    @observe('is_active')
    def _on_is_active_changed(self, *args):
        if self.disabled_msg:
            return

        # toggle visibility of markers
        for mark in self.marks.values():
            mark.visible = self.is_active

        # subscribe/unsubscribe to keypress events across all viewers
        for viewer in self.app._viewer_store.values():
            if not hasattr(viewer, 'figure'):
                # table viewer, etc
                continue
            callback = self._viewer_callback(viewer, self._on_viewer_key_event)

            if self.is_active:
                viewer.add_event_callback(callback, events=['keydown'])
            else:
                viewer.remove_event_callback(callback)

    # this is where items are being added to the table
    def _on_viewer_key_event(self, viewer, data):
        if data['event'] == 'keydown' and data['key'] == 'm':
            row_info = self.coords_info.as_dict()
            if 'viewer' in self.table.headers_avail:
                row_info['viewer'] = viewer.reference if viewer.reference is not None else viewer.reference_id  # noqa

            for k in self.table.headers_avail:
                row_info.setdefault(k, self._default_table_values.get(k, ''))

            try:
                self.table.add_item({k: v for k, v in row_info.items()
                                     if k in self.table.headers_avail})
            except ValueError as err:  # pragma: no cover
                raise ValueError(f'failed to add {row_info} to table: {repr(err)}')

            x, y = row_info['axes_x'], row_info['axes_y']
            self._get_mark(viewer).append_xy(getattr(x, 'value', x), getattr(y, 'value', y))

            self.hub.broadcast(MarkersPluginUpdate(table_length=len(self.table), sender=self))



import os

from contextlib import contextmanager
from traitlets import Bool, Dict, Unicode, Integer, List, observe

from jdaviz.core.template_mixin import (TemplateMixin, LayerSelect,
                                        LayerSelectMixin, DatasetSelectMixin)
from jdaviz.core.user_api import UserApiWrapper
from jdaviz.core.events import (IconsUpdatedMessage, AddDataMessage,
                                ChangeRefDataMessage, ViewerRenamedMessage)
from jdaviz.utils import cmap_samples, is_not_wcs_only

from glue.core.edit_subset_mode import (AndMode, AndNotMode, OrMode,
                                        ReplaceMode, XorMode, NewMode)
from glue.icons import icon_path
from glue_jupyter.common.toolbar_vuetify import read_icon

__all__ = ['DataMenu']


SUBSET_TOOL_IDS = {'circle': 'bqplot:truecircle',
                   'rectangle': 'bqplot:rectangle',
                   'ellipse': 'bqplot:ellipse',
                   'annulus': 'bqplot:circannulus',
                   'xrange': 'bqplot:xrange',
                   'yrange': 'bqplot:yrange'}

SUBSET_NAMES = {v: k for k, v in SUBSET_TOOL_IDS.items()}

SUBSET_MODES = {
    'new': NewMode,
    'replace': ReplaceMode,
    'or': OrMode,
    'and': AndMode,
    'xor': XorMode,
    'andnot': AndNotMode,
}


class DataMenu(TemplateMixin, LayerSelectMixin, DatasetSelectMixin):
    """Viewer Data Menu

    Only the following attributes and methods are available through the
    :ref:`public API <plugin-apis>`:

    * :meth:`open_menu`
    * :meth:`data_labels_loaded`
    * :meth:`data_labels_visible`
    * :meth:`data_labels_unloaded`
    * ``layer`` (:class:`~jdaviz.core.template_mixin.LayerSelect`):
        actively selected layer(s)
    * ``orientation`` (:class:`~jdaviz.core.template_mixin.LayerSelect`):
        orientation chosen for the viewer, if applicable
    * :meth:`set_layer_visibility`
    * :meth:`toggle_layer_visibility`
    * :meth:`create_subset`
    * :meth:`modify_subset`
    * :meth:`add_data`
    * :meth:`view_info`
    * :meth:`remove_from_viewer`
    * :meth:`remove_from_app`
    """
    template_file = __file__, "data_menu.vue"

    force_open_menu = Bool(False).tag(sync=True)

    viewer_id = Unicode().tag(sync=True)
    viewer_reference = Unicode().tag(sync=True)

    icons = Dict().tag(sync=True)
    layer_icons = Dict().tag(sync=True)  # read-only, see app.state.layer_icons
    viewer_icons = Dict().tag(sync=True)  # read-only, see app.state.viewer_icons

    visible_layers = Dict().tag(sync=True)  # read-only, set by viewer

    orientation_enabled = Bool(False).tag(sync=True)
    orientation_align_by_wcs = Bool(False).tag(sync=True)
    orientation_layer_items = List().tag(sync=True)
    orientation_layer_selected = Unicode().tag(sync=True)

    cmap_samples = Dict(cmap_samples).tag(sync=True)
    subset_tools = List().tag(sync=True)

    dm_layer_selected = List().tag(sync=True)

    loaded_n_data = Integer(0).tag(sync=True)
    selected_n_layers = Integer(0).tag(sync=True)
    selected_n_data = Integer(0).tag(sync=True)
    selected_n_subsets = Integer(0).tag(sync=True)

    info_enabled = Bool(False).tag(sync=True)
    info_tooltip = Unicode().tag(sync=True)

    delete_enabled = Bool(False).tag(sync=True)
    delete_tooltip = Unicode().tag(sync=True)
    delete_viewer_tooltip = Unicode().tag(sync=True)
    delete_app_enabled = Bool(False).tag(sync=True)
    delete_app_tooltip = Unicode().tag(sync=True)

    subset_edit_enabled = Bool(False).tag(sync=True)
    subset_edit_tooltip = Unicode().tag(sync=True)

    subset_edit_modes = List().tag(sync=True)

    def __init__(self, viewer, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._viewer = viewer
        self._during_select_sync = False

        # TODO: refactor how this is applied by default to go through filters directly
        self.layer.remove_filter('filter_is_root')
        self.layer.add_filter(is_not_wcs_only)
        self.layer.multiselect = True
        self.layer.sort_by = 'zorder'
        self.layer._default_mode = 'empty'

        # we'll use a modified version of the dataset mixin to have a filtered
        # list of data entries in the app that are not in the current viewer.
        # changing the selection has no consequence.
        def data_not_in_viewer(data):
            return data.label not in self.layer.choices
        self.dataset.filters = ['is_not_wcs_only', 'not_child_layer', data_not_in_viewer]

        self.orientation = LayerSelect(self,
                                       'orientation_layer_items',
                                       'orientation_layer_selected',
                                       'viewer_id',
                                       only_wcs_layers=True)
        self.orientation_enabled = self.config == 'imviz'

        # first attach callback to catch any updates to viewer/layer icons and then
        # set their initial state
        self.hub.subscribe(self, IconsUpdatedMessage, self._on_app_icons_updated)
        self.hub.subscribe(self, AddDataMessage, handler=lambda _: self._set_viewer_id())
        self.hub.subscribe(self, ChangeRefDataMessage, handler=self._on_refdata_change)
        self.hub.subscribe(self, ViewerRenamedMessage, handler=self._on_viewer_renamed_message)
        self.viewer_icons = dict(self.app.state.viewer_icons)
        self.layer_icons = dict(self.app.state.layer_icons)

        self.subset_edit_modes = [{'glue_name': 'replace', 'icon': read_icon(os.path.join(icon_path("glue_replace", icon_format="svg")), 'svg+xml')},  # noqa
                                  {'glue_name': 'or', 'icon': read_icon(os.path.join(icon_path("glue_or", icon_format="svg")), 'svg+xml')},  # noqa
                                  {'glue_name': 'and', 'icon': read_icon(os.path.join(icon_path("glue_and", icon_format="svg")), 'svg+xml')},  # noqa
                                  {'glue_name': 'xor', 'icon': read_icon(os.path.join(icon_path("glue_xor", icon_format="svg")), 'svg+xml')},  # noqa
                                  {'glue_name': 'andnot', 'icon': read_icon(os.path.join(icon_path("glue_andnot", icon_format="svg")), 'svg+xml')}]  # noqa

        # this currently assumes that toolbar.tools_data is set at init and does not change
        # if we ever support dynamic tool registration, this will need to be updated
        self.subset_tools = [{'id': k, 'img': v['img'], 'name': SUBSET_NAMES.get(k, k)}
                             for k, v in self._viewer.toolbar.tools_data.items()
                             if k in SUBSET_TOOL_IDS.values()]

        self.icons = {k: v for k, v in self.app.state.icons.items()}

    @property
    def user_api(self):
        expose = ['open_menu', 'layer', 'set_layer_visibility', 'toggle_layer_visibility',
                  'create_subset', 'modify_subset', 'add_data', 'view_info',
                  'remove_from_viewer', 'remove_from_app']
        readonly = ['data_labels_loaded', 'data_labels_visible', 'data_labels_unloaded']
        if self.app.config == 'imviz':
            expose += ['orientation']
        return UserApiWrapper(self, expose=expose, readonly=readonly)

    def open_menu(self):
        """
        Open all instances of the data menu.
        """
        self.force_open_menu = True

    @property
    def existing_subset_labels(self):
        return [sg.label for sg in self.app.data_collection.subset_groups]

    @property
    def data_labels_loaded(self):
        return [layer['label'] for layer in self.layer_items
                if layer['label'] not in self.existing_subset_labels]

    @property
    def data_labels_visible(self):
        return [layer['label'] for layer in self.layer_items
                if layer['label'] not in self.existing_subset_labels and layer['visible']]

    @property
    def data_labels_unloaded(self):
        return self.dataset.choices

    def _set_viewer_id(self):
        # viewer_ids are not populated on the viewer at init, so we'll keep checking and set
        # these the first time that they are available
        if len(self.viewer_id) and len(self.viewer_reference):
            return
        try:
            self.viewer_id = getattr(self._viewer, '_reference_id', '')
            self.viewer_reference = self._viewer.reference
            self.layer.viewer = self._viewer.reference
        except AttributeError:
            return

    def _on_app_icons_updated(self, msg):
        if msg.icon_type == 'viewer':
            self.viewer_icons = msg.icons
        elif msg.icon_type == 'layer':
            self.layer_icons = msg.icons
        self._set_viewer_id()

    def _on_refdata_change(self, msg):
        if msg.viewer_id != self.viewer_id:
            return
        self.orientation_align_by_wcs = self._viewer.state.reference_data.meta.get('_WCS_ONLY', False)  # noqa
        if self.orientation_align_by_wcs:
            with self.during_select_sync():
                self.orientation.selected = str(self._viewer.state.reference_data.label)

    def _on_viewer_renamed_message(self, msg):
        if self.viewer_reference == msg.old_viewer_ref:
            self.viewer_reference = msg.new_viewer_ref
            self._set_viewer_id()

    @observe('orientation_layer_selected')
    def _orientation_layer_selected_changed(self, event={}):
        if not hasattr(self, 'orientation'):
            return
        if self._during_select_sync:
            return
        op = self._viewer.jdaviz_helper.plugins['Orientation']
        op._obj.set_orientation_for_viewer(event.get('new'), self.viewer_id)

    @contextmanager
    def during_select_sync(self):
        self._during_select_sync = True
        try:
            yield
        except Exception:  # pragma: no cover
            self._during_select_sync = False
            raise
        self._during_select_sync = False

    @observe('dm_layer_selected', 'layer_multiselect')
    def _dm_layer_selected_changed(self, event={}):
        if not hasattr(self, 'layer'):  # pragma: no cover
            return
        if self._during_select_sync:
            return
        if event.get('name') == 'layer_multiselect' and event.get('new'):
            return
        if not self.layer.multiselect and len(self.dm_layer_selected) > 1:
            # vue will still treat the element as a list, so we will include the
            # logic here to enforce single-select toggling
            self.dm_layer_selected = [self.dm_layer_selected[-1]]
            return
        if (event.get('name') == 'dm_layer_selected'
                and len(event.get('new')) == len(event.get('old'))):
            # not possible from UI interaction, but instead caused by a selected
            # layer being removed (deleting a selected subset, etc).  We want
            # to update dm_layer_selected in order to preserve layer.selected
            self._layers_changed(event)
            return
        with self.during_select_sync():
            # map index in dm_layer_selected (inverse order of layer_items)
            # to set self.layer.selected
            selected = [self.layer_items[i]['label']
                        for i in self.dm_layer_selected]
            if self.layer.multiselect:
                self.layer.selected = selected
            else:
                self.layer.selected = selected[0] if len(selected) else ''

    @observe('layer_selected', 'layer_items')
    def _layers_changed(self, event={}):
        if not hasattr(self, 'layer') or not self.layer.multiselect:  # pragma: no cover
            return
        if not self._during_select_sync:
            with self.during_select_sync():
                # map list of strings in self.layer.selected to indices in dm_layer_selected
                layer_labels = [layer['label'] for layer in self.layer_items]
                layer_selected = self.layer_selected if self.layer.multiselect else [self.layer_selected]  # noqa
                self.dm_layer_selected = [layer_labels.index(label) for label in layer_selected
                                          if label in layer_labels]

        subset_labels = self.existing_subset_labels

        if event.get('name') == 'layer_items':
            # changing the layers in the viewer needs to trigger an update to dataset_items
            # through the set filters
            self.dataset._update_items()
            self.loaded_n_data = len([lyr for lyr in self.layer.choices
                                      if lyr not in subset_labels])
            return

        # update internal counts and tooltips
        self.selected_n_layers = len(self.layer.selected)
        self.selected_n_subsets = len([lyr for lyr in self.layer.selected if lyr in subset_labels])
        self.selected_n_data = self.selected_n_layers - self.selected_n_subsets

        # user-friendly representation of selection
        selected_repr = ""
        if self.selected_n_data:
            selected_repr += f"data ({self.selected_n_data})"
        if self.selected_n_subsets:
            if self.selected_n_data:
                selected_repr += " and"
            if self.selected_n_subsets == 1:
                selected_repr += f" subset ({self.selected_n_subsets})"
            else:
                selected_repr += f" subsets ({self.selected_n_subsets})"

        # layer info rules
        if self.selected_n_layers == 1:
            if max(self.dm_layer_selected) >= len(self.layer_items):  # pragma: no cover
                # can happen during state transition but should immediately be followed up
                # with an update
                self.info_enabled = False
                self.info_tooltip = ''
            elif self.layer_items[self.dm_layer_selected[0]].get('from_plugin', False):
                self.info_enabled = False
                self.info_tooltip = 'Selected data layer is a plugin product and does not have metadata'  # noqa
            else:
                self.info_enabled = True
                if self.selected_n_data == 1:
                    self.info_tooltip = 'View metadata for selected data'
                else:
                    self.info_tooltip = 'View subset info for selected subset'
        else:
            self.info_enabled = False
            if self.selected_n_layers == 0:
                self.info_tooltip = 'Select a layer to view info'
            else:
                self.info_tooltip = 'Select a single layer to view info'

        # delete layer rules
        if self.selected_n_layers == 0:
            self.delete_tooltip = "Select layer(s) to delete"
            self.delete_enabled = False
        else:
            self.delete_tooltip = f"Remove selected {selected_repr}"
            self.delete_enabled = True

        # delete from entire app rules
        subset_str = "subset" if self.selected_n_subsets == 1 else "subsets"
        if self.selected_n_subsets and self.selected_n_data:
            self.delete_viewer_tooltip = f"Remove selected data and hide selected {subset_str} in this viewer"  # noqa
        elif self.selected_n_data:
            self.delete_viewer_tooltip = "Remove selected data from this viewer"
        elif self.selected_n_subsets:
            self.delete_viewer_tooltip = f"Hide selected {subset_str} in this viewer"

        delete_app_tooltip = "Remove from all viewers and application (permanent, might affect existing subsets)"  # noqa
        if self.app.config == 'cubeviz':
            # forbid deleting non-plugin generated data
            selected_items = self.layer.selected_item
            for i, layer in enumerate(self.layer.selected):
                if (layer not in self.existing_subset_labels
                        and selected_items['from_plugin'][i] is None):
                    self.delete_app_enabled = False
                    self.delete_app_tooltip = f"Cannot delete imported data from {self.app.config}"
                    break
            else:
                self.delete_app_enabled = True
                self.delete_app_tooltip = delete_app_tooltip
        else:
            self.delete_app_enabled = True
            self.delete_app_tooltip = delete_app_tooltip

        # subset edit rules
        if self.selected_n_subsets == 1 and self.selected_n_layers == 1:
            self.subset_edit_enabled = True
            self.subset_edit_tooltip = f"Edit {self.layer_selected[0]}"
        else:
            self.subset_edit_enabled = False
            if self.selected_n_subsets == 0:
                self.subset_edit_tooltip = "Select a subset to edit"
            else:
                self.subset_edit_tooltip = "Select a single subset to edit"

    def set_layer_visibility(self, layer_label, visible=True):
        """
        Set the visibility of a layer in the viewer.

        Parameters
        ----------
        layer_label : str
            The label of the layer to set the visibility of.
        visible : bool
            Whether the layer should be visible or not.

        Returns
        -------
        dict
            A dictionary of the current visible layers.
        """
        for layer in self._viewer.layers:
            if layer.layer.label == layer_label:
                layer.visible = visible
            elif hasattr(layer.layer, 'data') and layer.layer.data.label == layer_label:
                layer.visible = layer.layer.label in self.visible_layers
            if not visible and self.app._get_assoc_data_parent(layer.layer.label) == layer_label:
                # then this is a child-layer of a parent-layer that is being hidden
                # so also hide the child-layer
                layer.visible = False

        if visible and (parent_label := self.app._get_assoc_data_parent(layer_label)):
            # ensure the parent layer is also visible
            self.set_layer_visibility(parent_label, visible=True)

        return self.visible_layers

    def toggle_layer_visibility(self, layer_label):
        """
        Toggle the visibility of a layer in the viewer.

        Parameters
        ----------
        layer_label : str
            The label of the layer to toggle the visibility of.

        Returns
        -------
        bool
            The new visibility state of the layer.
        """
        visible = layer_label not in self.visible_layers
        self.set_layer_visibility(layer_label, visible=visible)
        return visible

    def vue_set_layer_visibility(self, info, *args):
        return self.set_layer_visibility(info.get('layer'), info.get('value'))  # pragma: no cover

    def add_data(self, *data_labels):
        """
        Add a dataset to the viewer.

        Parameters
        ----------
        *data_labels : str
            The label(s) of the dataset to add to the viewer.
        """
        unavailable = [data_label for data_label in data_labels
                       if data_label not in self.dataset.choices]
        if len(unavailable):
            raise ValueError(f"Data labels {unavailable} not able to be loaded into '{self.viewer_id}'.  Must be one of: {self.dataset.choices}")  # noqa
        for data_label in data_labels:
            self.app.add_data_to_viewer(self.viewer_id, data_label)

    def vue_add_data_to_viewer(self, info, *args):
        self.add_data(info.get('data_label'))  # pragma: no cover

    def create_subset(self, subset_type):
        """
        Interactively create a new subset in the viewer.  This sets the app-wide subset
        selection to 'Create New' and selects the appropriate tool in this viewer's toolbar.

        Parameters
        ----------
        subset_type : str
            The type of subset to create.  Must be one of 'circle', 'rectangle', 'ellipse',
            'annulus', 'xrange', or 'yrange', and must be an available tool in this viewer.
        """
        # clear previous selection, finalize subsets, temporarily sets default tool
        self._viewer.toolbar.active_tool_id = None
        # set toolbar to the selection, will also set app-wide subset selection to "Create New"
        # NOTE: supports passing either the user-friendly name or the actual ID
        self._viewer.toolbar.select_tool(SUBSET_TOOL_IDS.get(subset_type, subset_type))

    def vue_create_subset(self, info, *args):
        self.create_subset(info.get('subset_type'))  # pragma: no cover

    def modify_subset(self, combination_mode, subset_type):
        """
        Interactively modify an existing subset in the viewer.  This sets the app-wide subset
        selection to the currently selected subset, mode to the selected combination_mode,
        and selects the appropriate tool in this viewer's toolbar.

        Parameters
        ----------
        combination_mode : str
            The combination mode to apply to the subset.  Must be one of 'replace', 'or', 'and',
            'xor', or 'andnot'.
        subset_type : str
            The type of subset to modify.  Must be one of 'circle', 'rectangle', 'ellipse',
            'annulus', 'xrange', or 'yrange', and must be an available tool in this viewer.
        """
        # future improvement: allow overriding layer.selected, with pre-validation
        if len(self.layer.selected) != 1:
            raise ValueError("Only one layer can be selected to modify subset.")
        if self.layer.selected[0] not in self.existing_subset_labels:
            raise ValueError("Selected layer is not a subset.")
        subset = self.layer.selected[0]

        # set tool first since that might default to "Create New"
        self._viewer.toolbar.select_tool(SUBSET_TOOL_IDS.get(subset_type, subset_type))
        # set subset selection to the subset to modify
        subset_grp = [sg for sg in self.app.data_collection.subset_groups if sg.label == subset]
        self.session.edit_subset_mode.edit_subset = subset_grp
        # set combination mode
        self.session.edit_subset_mode.mode = SUBSET_MODES.get(combination_mode)

    def vue_modify_subset(self, info, *args):
        self.modify_subset(info.get('combination_mode'),
                           info.get('subset_type'))  # pragma: no cover

    def view_info(self):
        """
        View info for the selected layer by opening either the metadata or subset plugin to the
        selected entry.
        """
        # future improvement: allow overriding layer.selected, with pre-validation
        if len(self.layer.selected) != 1:
            raise ValueError("Only one layer can be selected to view info.")
        if self.layer.selected[0] in self.existing_subset_labels:
            sp = self._viewer.jdaviz_helper.plugins.get('Subset Tools', None)
            if sp is None:  # pragma: no cover
                raise ValueError("subset tools plugin not available")
            sp._obj.subset.selected = self.layer.selected[0]
            sp.open_in_tray()
        else:
            mp = self._viewer.jdaviz_helper.plugins.get('Metadata', None)
            if mp is None:  # pragma: no cover
                raise ValueError("metadata plugin not available")
            mp.dataset.selected = self.layer.selected[0]
            mp.open_in_tray()

    def vue_view_info(self, *args):
        self.view_info()  # pragma: no cover

    def remove_from_viewer(self):
        """
        Remove the selected layers from the viewer.  For subset layers, this
        sets the visibility of the subset layer.  For data layers,
        this unloads the data from the viewer, but keeps it in the application or other viewers.
        """
        # future improvement: allow overriding layer.selected via *args, with pre-validation
        for layer in self.layer.selected:
            if layer in self.existing_subset_labels:
                self.set_layer_visibility(layer, visible=False)
            else:
                self.app.remove_data_from_viewer(self.viewer_id, layer)

    def vue_remove_from_viewer(self, *args):
        self.remove_from_viewer()  # pragma: no cover

    def remove_from_app(self):
        """
        Remove the selected layers from the entire app and all viewers.
        """
        # future improvement: allow overriding layer.selected via *args, with pre-validation
        for layer in self.layer.selected:
            if layer in self.existing_subset_labels:
                for sg in self.app.data_collection.subset_groups:
                    if sg.label == layer:
                        self.app.data_collection.remove_subset_group(sg)
                        break
            else:
                self.app.data_item_remove(layer)

    def vue_remove_from_app(self, *args):
        self.remove_from_app()  # pragma: no cover

    def vue_open_orientation_plugin(self, *args):
        op = self._viewer.jdaviz_helper.plugins.get('Orientation', None)
        if op is None:  # pragma: no cover
            raise ValueError("orientation plugin not available")
        op.open_in_tray()



from astropy.io.fits import PrimaryHDU
from ..plugins.parsers import _get_source_identifiers_by_hdu, FALLBACK_NAME


def test_SOURCEID():
    hdu = PrimaryHDU()
    hdu.header['SOURCEID'] = 'Target 1 SOURCEID'
    assert _get_source_identifiers_by_hdu([hdu]) == ['Target 1 SOURCEID']

    hdu2 = PrimaryHDU()
    hdu2.header['SOURCEID'] = 'Target 2 SOURCEID'
    assert _get_source_identifiers_by_hdu([hdu, hdu2]) == ['Target 1 SOURCEID', 'Target 2 SOURCEID']


def test_OBJECT():
    hdu = PrimaryHDU()
    hdu.header['OBJECT'] = 'Target 1 OBJECT'
    assert _get_source_identifiers_by_hdu([hdu]) == ['Target 1 OBJECT']

    hdu2 = PrimaryHDU()
    hdu2.header['OBJECT'] = 'Target 2 OBJECT'
    assert _get_source_identifiers_by_hdu([hdu, hdu2]) == ['Target 1 OBJECT', 'Target 2 OBJECT']


def test_priority():
    hdu = PrimaryHDU()
    hdu.header['SOURCEID'] = 'Target 1 SOURCEID'
    hdu.header['OBJECT'] = 'Target 1 OBJECT'
    assert _get_source_identifiers_by_hdu([hdu]) == ['Target 1 SOURCEID']

    hdu2 = PrimaryHDU()
    hdu2.header['SOURCEID'] = 'Target 2 SOURCEID'
    hdu2.header['OBJECT'] = 'Target 2 OBJECT'
    assert _get_source_identifiers_by_hdu([hdu, hdu2]) == ['Target 1 SOURCEID', 'Target 2 SOURCEID']


def test_priority_mixed():
    hdu = PrimaryHDU()
    hdu.header['SOURCEID'] = 'Target 1 SOURCEID'
    hdu.header['OBJECT'] = 'Target 1 OBJECT'
    assert _get_source_identifiers_by_hdu([hdu]) == ['Target 1 SOURCEID']

    hdu2 = PrimaryHDU()
    hdu2.header['OBJECT'] = 'Target 2 OBJECT'
    assert _get_source_identifiers_by_hdu([hdu, hdu2]) == ['Target 1 SOURCEID', 'Target 2 OBJECT']


def test_empty_fallback():
    hdu = PrimaryHDU()
    assert _get_source_identifiers_by_hdu([hdu]) == [FALLBACK_NAME]
    assert _get_source_identifiers_by_hdu([hdu, hdu]) == [FALLBACK_NAME, FALLBACK_NAME]


def test_single_filename_fallback():
    hdu = PrimaryHDU()
    assert _get_source_identifiers_by_hdu([hdu], "/path/to/test_file.fits") == ['test_file.fits']
    assert _get_source_identifiers_by_hdu([hdu, hdu],
                                          "/path/to/test_file.fits"
                                          ) == ['test_file.fits', 'test_file.fits']


def test_filename_fallback():
    hdu = PrimaryHDU()
    assert _get_source_identifiers_by_hdu([hdu], ["/path/to/test_file.fits"]) == ['test_file.fits']
    assert _get_source_identifiers_by_hdu([hdu, hdu],
                                          ["/path/to/test_file1.fits", "/path/to/test_file2.fits"]
                                          ) == ['test_file1.fits', 'test_file2.fits']


def test_custom_header():
    hdu = PrimaryHDU()
    hdu.header['CUSTOM'] = "Target 1 Custom"
    assert _get_source_identifiers_by_hdu([hdu], header_keys='CUSTOM') == ["Target 1 Custom"]


def test_custom_header_priority_mixed():
    hdu = PrimaryHDU()
    hdu.header['CUSTOM'] = "Target 1 CUSTOM"
    assert _get_source_identifiers_by_hdu([hdu],
                                          header_keys=['CUSTOM', 'CUSTOM2']
                                          ) == ["Target 1 CUSTOM"]

    hdu2 = PrimaryHDU()
    hdu2.header['CUSTOM2'] = "Target 2 CUSTOM2"
    assert _get_source_identifiers_by_hdu([hdu, hdu2],
                                          header_keys=['CUSTOM', 'CUSTOM2']
                                          ) == ["Target 1 CUSTOM", "Target 2 CUSTOM2"]



import warnings

import pytest

from jdaviz.configs.cubeviz.plugins.slice.slice import Slice


def test_slice(cubeviz_helper, spectrum1d_cube):
    app = cubeviz_helper.app
    sl = Slice(app=app)

    # No data yet
    assert len(sl.slice_selection_viewers) == 2  # flux-viewer, uncert-viewer
    assert len(sl.slice_indicator_viewers) == 1  # spectrum-viewer
    assert len(sl.valid_indicator_values_sorted) == 0
    assert len(sl.valid_selection_values_sorted) == 0

    # Make sure nothing crashes if plugin used without data]
    sl.vue_play_next()
    sl.vue_play_start_stop()
    assert not sl.is_playing

    cubeviz_helper.load_data(spectrum1d_cube, data_label='test')
    app.add_data_to_viewer("flux-viewer", "test[FLUX]")
    app.add_data_to_viewer("uncert-viewer", "test[FLUX]")
    app.add_data_to_viewer("spectrum-viewer", "Spectrum (sum)")
    sv = cubeviz_helper.viewers['spectrum-viewer']._obj

    # sample cube only has 2 slices with wavelengths [4.62280007e-07 4.62360028e-07] m
    assert len(sv.slice_values) == 2
    assert len(sl.valid_indicator_values_sorted) == 2
    slice_values = sl.valid_selection_values_sorted
    assert len(slice_values) == 2

    assert sl.value == slice_values[1]
    assert cubeviz_helper.app.get_viewer("flux-viewer").slice == 1
    assert cubeviz_helper.app.get_viewer("flux-viewer").state.slices[-1] == 1
    assert cubeviz_helper.app.get_viewer("uncert-viewer").state.slices[-1] == 1
    sl.value = slice_values[0]
    assert cubeviz_helper.app.get_viewer("flux-viewer").slice == 0
    assert sl.value == slice_values[0]

    sl.value = slice_values[1]
    assert sl.value == slice_values[1]

    # test setting a static 2d image to the "watched" flux viewer to make sure it disconnects
    mm = app.get_tray_item_from_name('cubeviz-moment-maps')
    mm.add_to_viewer_selected = 'flux-viewer'
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', message=r'.*No observer defined on WCS.*')
        mm.vue_calculate_moment()

    # test in conjunction with as_steps
    sv = app.get_viewer('spectrum-viewer')
    orig_len = len(sv.native_marks[0].x)

    sv.state.layers[0].as_steps = True
    new_len = len(sv.native_marks[0].x)
    assert new_len == 2*orig_len
    sl.value = 4.62360028e-07
    assert sl.value == slice_values[1]

    # Add test for unit conversion
    uc_plugin = cubeviz_helper.plugins['Unit Conversion']._obj
    uc_plugin.spectral_unit.selected = 'Angstrom'
    assert sl.value_unit == 'Angstrom'
    sl.value = 4623.60028
    assert sl.value == 4623.600276968349

    # Retrieve updated slice_values
    slice_values = sl.valid_selection_values_sorted

    # Test player buttons API

    sl.vue_goto_first()
    assert sl.value == slice_values[0]

    sl.vue_goto_last()
    assert sl.value == slice_values[-1]

    sl.vue_play_next()  # Should automatically wrap to beginning
    assert sl.value == slice_values[0]

    sl.vue_play_start_stop()  # Start
    assert sl.is_playing
    assert sl._player.is_alive()
    sl.vue_play_next()  # Should be no-op
    sl.vue_goto_last()  # Should be no-op
    sl.vue_goto_first()  # Should be no-op
    sl.vue_play_start_stop()  # Stop
    assert not sl.is_playing
    assert not sl._player
    # NOTE: Hard to check sl.slice here because it is non-deterministic.


def test_indicator_settings(cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube, data_label='test')
    app = cubeviz_helper.app
    app.add_data_to_viewer("flux-viewer", "test[FLUX]")
    app.add_data_to_viewer("spectrum-viewer", "Spectrum (sum)")
    sl = cubeviz_helper.plugins['Slice']._obj
    sv = app.get_viewer('spectrum-viewer')
    indicator = sv.slice_indicator

    assert sl.show_indicator is True
    assert indicator._show_if_inactive is True
    assert sl.show_value is True
    assert indicator.label.visible is True

    sl.show_indicator = False
    assert indicator._show_if_inactive is False

    sl.show_value = False
    assert indicator.label.visible is False


@pytest.mark.filterwarnings('ignore:No observer defined on WCS')
def test_init_slice(cubeviz_helper, spectrum1d_cube):
    cubeviz_helper.load_data(spectrum1d_cube, data_label='test')

    fv = cubeviz_helper.app.get_viewer('flux-viewer')
    sl = cubeviz_helper.plugins['Slice']
    slice_values = sl._obj.valid_selection_values_sorted

    assert sl.value == slice_values[1]
    assert fv.slice == 1
    assert fv.state.slices == (0, 0, 1)

    # make sure adding new data doesn't revert slice to 0
    mm = cubeviz_helper.plugins['Moment Maps']
    mm.calculate_moment(add_data=True)

    assert sl.value == slice_values[1]
    assert fv.slice == 1
    assert fv.state.slices == (0, 0, 1)





def test_load_rectangular_ramp(rampviz_helper, jwst_level_1b_rectangular_ramp):
    rampviz_helper.load_data(jwst_level_1b_rectangular_ramp)

    # drop the integration axis
    original_cube_shape = jwst_level_1b_rectangular_ramp.shape[1:]

    # on ramp cube load (1), the parser loads a diff cube (2) and
    # the ramp extraction plugin produces a default extraction (3):
    assert len(rampviz_helper.app.data_collection) == 3

    parsed_cube_shape = rampviz_helper.app.data_collection[0].shape
    assert parsed_cube_shape == (
        original_cube_shape[1], original_cube_shape[2], original_cube_shape[0]
    )


def test_load_level_1_and_2(
        rampviz_helper,
        jwst_level_1b_rectangular_ramp,
        jwst_level_2c_rate_image
):
    # load level 1 ramp and level 2 rate image
    rampviz_helper.load_data(jwst_level_1b_rectangular_ramp)
    rampviz_helper.load_data(jwst_level_2c_rate_image)

    # confirm that a "level-2" viewer is created, and that
    # the rate image is loaded into it
    assert len(rampviz_helper.viewers) == 4
    assert 'level-2' in rampviz_helper.viewers

    layers = rampviz_helper.app.get_viewer('level-2').layers
    assert len(layers) == 1
